async def _asyncWrapWith(res, wrapper_fn):
    result = await res
    return wrapper_fn(result["id"])


def wrapWith(res, wrapper_fn):
    if isinstance(res, dict):
        return wrapper_fn(res)
    else:
        return _asyncWrapWith(res, wrapper_fn)




from aiodataloader import DataLoader


class BooleanVariableLoader(DataLoader):
    def __init__(self, client, id):
        super().__init__()
        self.client = client
        self._id = id

    async def batch_load_fn(self, keys):
        fields = " ".join(set(keys))
        res = await self.client.query('{booleanVariable(id:"%s"){%s}}' % (self._id, fields), keys=["booleanVariable"])

        # if fetching object the key will be the first part of the field
        # e.g. when fetching thing{id} the result is in the thing key
        resolvedValues = [res[key.split("{")[0]] for key in keys]

        return resolvedValues


class BooleanVariable:
    def __init__(self, client, id):
        self.client = client
        self._id = id
        self.loader = BooleanVariableLoader(client, id)

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        if self.client.asyncio:
            return self.loader.load("name")
        else:
            return self.client.query('{booleanVariable(id:"%s"){name}}' % self._id, keys=[
                "booleanVariable", "name"])

    @name.setter
    def name(self, newName):
        self.client.mutation(
            'mutation{booleanVariable(id:"%s", name:"%s"){id}}' % (self._id, newName), asyncio=False)

    @property
    def developer_only(self):
        if self.client.asyncio:
            return self.loader.load("developerOnly")
        else:
            return self.client.query('{booleanVariable(id:"%s"){developerOnly}}' % self._id, keys=[
                "booleanVariable", "developerOnly"])

    @developer_only.setter
    def developer_only(self, newValue):
        self.client.mutation(
            'mutation{booleanVariable(id:"%s", developerOnly:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def permission(self):
        if self.client.asyncio:
            return self.loader.load("permission")
        else:
            return self.client.query('{booleanVariable(id:"%s"){permission}}' % self._id, keys=[
                "booleanVariable", "permission"])

    @permission.setter
    def permission(self, newValue):
        self.client.mutation(
            'mutation{booleanVariable(id:"%s", permission:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def hidden(self):
        if self.client.asyncio:
            return self.loader.load("hidden")
        else:
            return self.client.query('{booleanVariable(id:"%s"){hidden}}' % self._id, keys=[
                "booleanVariable", "hidden"])

    @hidden.setter
    def hidden(self, newValue):
        self.client.mutation(
            'mutation{booleanVariable(id:"%s", hidden:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def index(self):
        if self.client.asyncio:
            return self.loader.load("index")
        else:
            return self.client.query('{booleanVariable(id:"%s"){index}}' % self._id, keys=[
                "booleanVariable", "index"])

    @index.setter
    def index(self, newValue):
        self.client.mutation(
            'mutation{booleanVariable(id:"%s", index:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def my_role(self):
        if self.client.asyncio:
            return self.loader.load("myRole")
        else:
            return self.client.query('{booleanVariable(id:"%s"){myRole}}' % self._id, keys=[
                "booleanVariable", "myRole"])

    @property
    def created_at(self):
        if self.client.asyncio:
            return self.loader.load("createdAt")
        else:
            return self.client.query('{booleanVariable(id:"%s"){createdAt}}' % self._id, keys=[
                "booleanVariable", "createdAt"])

    @property
    def updated_at(self):
        if self.client.asyncio:
            return self.loader.load("updatedAt")
        else:
            return self.client.query('{booleanVariable(id:"%s"){updatedAt}}' % self._id, keys=[
                "booleanVariable", "updatedAt"])

    async def _async_load_thing(self):
        id = await self.loader.load("thing{id}")["id"]
        from .thing import Thing
        return Thing(self.client, id)

    @property
    def thing(self):
        if self.client.asyncio:
            return self._async_load_thing()
        else:
            id = self.client.query('{booleanVariable(id:"%s"){thing{id}}}' % self._id, keys=[
                "booleanVariable", "thing", "id"])

            from .thing import Thing
            return Thing(self.client, id)

    @property
    def value(self):
        if self.client.asyncio:
            return self.loader.load("value")
        else:
            return self.client.query('{booleanVariable(id:"%s"){value}}' % self._id, keys=[
                "booleanVariable", "value"])

    @value.setter
    def value(self, newValue):
        self.client.mutation(
            'mutation{booleanVariable(id:"%s", value:%s){id}}' % (self._id, "true" if newValue else "false"), asyncio=False)



from aiodataloader import DataLoader
from igloo.models.utils import wrapWith
from igloo.utils import get_representation


class NotificationLoader(DataLoader):
    def __init__(self, client, id):
        super().__init__()
        self.client = client
        self._id = id

    async def batch_load_fn(self, keys):
        fields = " ".join(set(keys))
        res = await self.client.query('{notification(id:"%s"){%s}}' % (self._id, fields), keys=["thing"])

        resolvedValues = [res[key] for key in keys]

        return resolvedValues


class Notification:
    def __init__(self, client, id):
        self.client = client
        self._id = id
        self.loader = NotificationLoader(client, id)

    @property
    def id(self):
        return self._id

    @property
    def createdAt(self):
        if self.client.asyncio:
            return self.loader.load("createdAt")
        else:
            return self.client.query('{notification(id:"%s"){createdAt}}' % self._id, keys=[
                "notification", "createdAt"])

    @property
    def updatedAt(self):
        if self.client.asyncio:
            return self.loader.load("updatedAt")
        else:
            return self.client.query('{notification(id:"%s"){updatedAt}}' % self._id, keys=[
                "notification", "updatedAt"])

    @property
    def thing(self):
        if self.client.asyncio:
            res = self.loader.load("thing{id}")
        else:
            res = self.client.query('{notification(id:"%s"){thing{id}}}' % self._id, keys=[
                "notification", "thing"])

        def wrapper(res):
            from .thing import Thing
            return Thing(self.client, res["id"])

        return wrapWith(res, wrapper)

    @property
    def content(self):
        if self.client.asyncio:
            return self.loader.load("content")
        else:
            return self.client.query('{notification(id:"%s"){content}}' %
                                     self._id, keys=["notification", "content"])

    @content.setter
    def content(self, newContent):
        self.client.mutation(
            'mutation{notification(id:"%s", content:"%s"){id}}' % (self._id, newContent), asyncio=False)

    @property
    def date(self):
        if self.client.asyncio:
            return self.loader.load("date")
        else:
            return self.client.query('{notification(id:"%s"){date}}' %
                                     self._id, keys=["notification", "date"])

    @property
    def read(self):
        if self.client.asyncio:
            return self.loader.load("read")
        else:
            return self.client.query('{notification(id:"%s"){read}}' %
                                     self._id, keys=["notification", "read"])

    @read.setter
    def read(self, newContent):
        self.client.mutation(
            'mutation{notification(id:"%s", read:"%s"){id}}' % (self._id, newContent), asyncio=False)


class ThingNotificationList:
    def __init__(self, client, thingId):
        self.client = client
        self.thingId = thingId
        self.current = 0
        self._filter = "{}"

    def filter(self, _filter):
        self._filter = get_representation(_filter)
        return self

    def __len__(self):
        res = self.client.query(
            '{thing(id:"%s"){notificationCount(filter:%s)}}' % (self.thingId, self._filter))
        return res["thing"]["notificationCount"]

    def __getitem__(self, i):
        if isinstance(i, int):
            res = self.client.query(
                '{thing(id:"%s"){notifications(limit:1, offset:%d, filter:%s){id}}}' % (self.thingId, i, self._filter))
            if len(res["thing"]["notifications"]) != 1:
                raise IndexError()
            return Notification(self.client, res["thing"]["notifications"][0]["id"])
        elif isinstance(i, slice):
            start, end, _ = i.indices(len(self))
            res = self.client.query(
                '{thing(id:"%s"){notifications(offset:%d, limit:%d, filter:%s){id}}}' % (self.thingId, start, end-start, self._filter))
            return [Notification(self.client, notification["id"]) for notification in res["thing"]["notifications"]]
        else:
            raise TypeError("Unexpected type {} passed as index".format(i))

    def __iter__(self):
        return self

    def __next__(self):
        res = self.client.query(
            '{thing(id:"%s"){notifications(limit:1, offset:%d, filter:%s){id}}}' % (self.thingId, self.current, self._filter))

        if len(res["thing", "notifications"]) != 1:
            raise StopIteration

        self.current += 1
        return Notification(self.client, res["thing"]["notifications"][0]["id"])

    def next(self):
        return self.__next__()



from .float_variable import FloatVariable
from .file_variable import FileVariable
from .boolean_variable import BooleanVariable
from .string_variable import StringVariable
from .float_series_variable import FloatSeriesVariable
from .category_series_variable import CategorySeriesVariable
from igloo.utils import get_representation


def Variable(client, id, resolveType):
    if resolveType == "FloatVariable":
        return FloatVariable(client, id)
    elif resolveType == "FileVariable":
        return FileVariable(client, id)
    elif resolveType == "BooleanVariable":
        return BooleanVariable(client, id)
    elif resolveType == "StringVariable":
        return StringVariable(client, id)
    elif resolveType == "FloatSeriesVariable":
        return FloatSeriesVariable(client, id)
    elif resolveType == "CategorySeriesVariable":
        return CategorySeriesVariable(client, id)


class ThingVariablesList:
    def __init__(self, client, thingId):
        self.client = client
        self.thingId = thingId
        self.current = 0
        self._filter = "{}"

    def filter(self, _filter):
        self._filter = get_representation(_filter)
        return self

    def __len__(self):
        res = self.client.query(
            '{thing(id:"%s"){valueCount(filter:%s)}}' % (self.thingId, self._filter))
        return res["thing"]["valueCount"]

    def __getitem__(self, i):
        if isinstance(i, int):
            res = self.client.query(
                '{thing(id:"%s"){values(limit:1, offset:%d, filter:%s){id __typename}}}' % (self.thingId, i, self._filter))
            if len(res["thing"]["values"]) != 1:
                raise IndexError()
            return Value(self.client, res["thing"]["values"][0]["id"], res["thing"]["values"][0]["__typename"])
        elif isinstance(i, slice):
            start, end, _ = i.indices(len(self))
            res = self.client.query(
                '{thing(id:"%s"){values(offset:%d, limit:%d, filter:%s){id __typename}}}' % (self.thingId, start, end-start, self._filter))
            return [Value(self.client, value["id"], value["__typename"]) for value in res["thing"]["values"]]
        else:
            raise TypeError("Unexpected type {} passed as index".format(i))

    def __iter__(self):
        return self

    def __next__(self):
        res = self.client.query(
            '{thing(id:"%s"){values(limit:1, offset:%d, filter:%s){id __typename}}}' % (self.thingId, self.current, self._filter))

        if len(res["thing", "values"]) != 1:
            raise StopIteration

        self.current += 1
        return Value(self.client, res["thing"]["values"][0]["id"], res["thing"]["values"][0]["__typename"])

    def next(self):
        return self.__next__()



from igloo.models.user import User
from igloo.models.variable import Variable
from igloo.models.access_token import AccessToken
from igloo.models.pending_share import PendingShare
from igloo.models.environment import Environment
from igloo.models.thing import Thing
from igloo.models.float_variable import FloatVariable
from igloo.models.pending_transfer import PendingTransfer
from igloo.models.notification import Notification
from igloo.models.boolean_variable import BooleanVariable
from igloo.models.string_variable import StringVariable
from igloo.models.float_series_variable import FloatSeriesVariable
from igloo.models.category_series_variable import CategorySeriesVariable
from igloo.models.category_series_node import CategorySeriesNode
from igloo.models.file_variable import FileVariable
from igloo.models.float_series_node import FloatSeriesNode
from igloo.main import _Undefined, UNDEFINED, undefined, GraphQLException, Client




from aiodataloader import DataLoader
from igloo.models.utils import wrapWith
from igloo.utils import get_representation


class ThingLoader(DataLoader):
    def __init__(self, client, id):
        super().__init__()
        self.client = client
        self._id = id

    async def batch_load_fn(self, keys):
        fields = " ".join(set(keys))
        res = await self.client.query('{thing(id:"%s"){%s}}' % (self._id, fields), keys=["thing"])

        resolvedValues = [res[key.split("{")[0]] for key in keys]

        return resolvedValues


class Thing:
    def __init__(self, client, id=None):
        self.client = client

        if id is None:
            self._id = self.client.query(
                '{thing{id}}', keys=["thing", "id"], asyncio=False)
        else:
            self._id = id

        self.loader = ThingLoader(client, self._id)

    @property
    def id(self):
        return self._id

    @property
    def createdAt(self):
        if self.client.asyncio:
            return self.loader.load("createdAt")
        else:
            return self.client.query('{thing(id:"%s"){createdAt}}' % self._id, keys=[
                "thing", "createdAt"])

    @property
    def updatedAt(self):
        if self.client.asyncio:
            return self.loader.load("updatedAt")
        else:
            return self.client.query('{thing(id:"%s"){updatedAt}}' % self._id, keys=[
                "thing", "updatedAt"])

    @property
    def type(self):
        if self.client.asyncio:
            return self.loader.load("type")
        else:
            return self.client.query('{thing(id:"%s"){type}}' %
                                     self._id, keys=["thing", "type"])

    @type.setter
    def type(self, newThingType):
        self.client.mutation(
            'mutation{thing(id:"%s", type:"%s"){id}}' % (self._id, newThingType), asyncio=False)

    @property
    def my_role(self):
        if self.client.asyncio:
            return self.loader.load("myRole")
        else:
            return self.client.query('{thing(id:"%s"){myRole}}' %
                                     self._id, keys=["thing", "myRole"])

    @property
    def starred(self):
        if self.client.asyncio:
            return self.loader.load("starred")
        else:
            return self.client.query('{thing(id:"%s"){starred}}' %
                                     self._id, keys=["thing", "starred"])

    @starred.setter
    def starred(self, newValue):
        self.client.mutation(
            'mutation{thing(id:"%s", starred:%s){id}}' % (self._id, "true" if newValue else "false"), asyncio=False)

    @property
    def name(self):
        if self.client.asyncio:
            return self.loader.load("name")
        else:
            return self.client.query('{thing(id:"%s"){name}}' %
                                     self._id, keys=["thing", "name"])

    @name.setter
    def name(self, newName):
        self.client.mutation(
            'mutation{thing(id:"%s", name:"%s"){id}}' % (self._id, newName), asyncio=False)

    @property
    def index(self):
        if self.client.asyncio:
            return self.loader.load("index")
        else:
            return self.client.query('{thing(id:"%s"){index}}' %
                                     self._id, keys=["thing", "index"])

    @index.setter
    def index(self, newValue):
        self.client.mutation(
            'mutation{thing(id:"%s", index:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def online(self):
        if self.client.asyncio:
            return self.loader.load("online")
        else:
            return self.client.query('{thing(id:"%s"){online}}' %
                                     self._id, keys=["thing", "online"])

    @online.setter
    def online(self, newValue):
        raise Exception(
            "You cannot set `online` directly, use the `keep_online` subscription")

    @property
    def token(self):
        if self.client.asyncio:
            return self.loader.load("token")
        else:
            return self.client.query('{thing(id:"%s"){token}}' %
                                     self._id, keys=["thing", "token"])

    @property
    def used_storage(self):
        if self.client.asyncio:
            return self.loader.load("usedStorage")
        else:
            return self.client.query('{thing(id:"%s"){usedStorage}}' %
                                     self._id, keys=["thing", "usedStorage"])

    @property
    def stored_notifications(self):
        if self.client.asyncio:
            return self.loader.load("storedNotifications")
        else:
            return self.client.query('{thing(id:"%s"){storedNotifications}}' %
                                     self._id, keys=["thing", "storedNotifications"])

    @stored_notifications.setter
    def stored_notifications(self, newValue):
        self.client.mutation(
            'mutation{thing(id:"%s", storedNotifications:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def signal_status(self):
        if self.client.asyncio:
            return self.loader.load("signalStatus")
        else:
            return self.client.query('{thing(id:"%s"){signalStatus}}' %
                                     self._id, keys=["thing", "signalStatus"])

    @signal_status.setter
    def signal_status(self, newValue):
        self.client.mutation(
            'mutation{thing(id:"%s", signalStatus:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def battery_status(self):
        if self.client.asyncio:
            return self.loader.load("batteryStatus")
        else:
            return self.client.query('{thing(id:"%s"){batteryStatus}}' %
                                     self._id, keys=["thing", "batteryStatus"])

    @battery_status.setter
    def battery_status(self, newValue):
        self.client.mutation(
            'mutation{thing(id:"%s", batteryStatus:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def battery_charging(self):
        if self.client.asyncio:
            return self.loader.load("batteryCharging")
        else:
            return self.client.query('{thing(id:"%s"){batteryCharging}}' %
                                     self._id, keys=["thing", "batteryCharging"])

    @battery_charging.setter
    def battery_charging(self, newValue):
        self.client.mutation(
            'mutation{thing(id:"%s", batteryCharging:%s){id}}' % (self._id, "true" if newValue else "false"), asyncio=False)

    @property
    def battery_threshold(self):
        if self.client.asyncio:
            return self.loader.load("batteryThreshold")
        else:
            return self.client.query('{thing(id:"%s"){batteryThreshold}}' %
                                     self._id, keys=["thing", "batteryThreshold"])

    @battery_threshold.setter
    def battery_threshold(self, newValue):
        self.client.mutation(
            'mutation{thing(id:"%s", batteryThreshold:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def firmware(self):
        if self.client.asyncio:
            return self.loader.load("firmware")
        else:
            return self.client.query('{thing(id:"%s"){firmware}}' %
                                     self._id, keys=["thing", "firmware"])

    @firmware.setter
    def firmware(self, newValue):
        self.client.mutation(
            'mutation{thing(id:"%s", firmware:"%s"){id}}' % (self._id, newValue), asyncio=False)

    @property
    def muted(self):
        if self.client.asyncio:
            return self.loader.load("muted")
        else:
            return self.client.query('{thing(id:"%s"){muted}}' %
                                     self._id, keys=["thing", "muted"])

    @muted.setter
    def muted(self, newValue):
        self.client.mutation(
            'mutation{thing(id:"%s", muted:%s){id}}' % (self._id, "true" if newValue else "false"), asyncio=False)

    @property
    def qr_code(self):
        if self.client.asyncio:
            return self.loader.load("qrCode")
        else:
            return self.client.query('{thing(id:"%s"){qrCode}}' %
                                     self._id, keys=["thing", "qrCode"])

    @property
    def claim_code(self):
        if self.client.asyncio:
            return self.loader.load("claimCode")
        else:
            return self.client.query('{thing(id:"%s"){claimCode}}' %
                                     self._id, keys=["thing", "claimCode"])

    @property
    def claimed(self):
        if self.client.asyncio:
            return self.loader.load("claimed")
        else:
            return self.client.query('{thing(id:"%s"){claimed}}' %
                                     self._id, keys=["thing", "claimed"])

    @property
    def environment(self):
        from .environment import Environment

        if self.client.asyncio:
            res = self.loader.load("environment{id}")
        else:
            res = self.client.query('{thing(id:"%s"){environment{id}}}' %
                                    self._id, keys=["thing", "environment"])

        def wrapper(res):
            return Environment(self.client, res["id"])

        return wrapWith(res, wrapper)

    @property
    def producer(self):
        from .user import User

        if self.client.asyncio:
            res = self.loader.load("producer{id}")
        else:
            res = self.client.query('{thing(id:"%s"){producer{id}}}' %
                                    self._id, keys=["thing", "producer"])

        def wrapper(res):
            return User(self.client, res["id"])

        return wrapWith(res, wrapper)

    @property
    def notifications(self):
        from .notification import ThingNotificationList
        return ThingNotificationList(self.client, self.id)

    @property
    def last_notification(self):
        from .notification import Notification

        if self.client.asyncio:
            res = self.loader.load("lastNotification{id}")
        else:
            res = self.client.query('{thing(id:"%s"){lastNotification{id}}}' %
                                    self._id, keys=["thing", "lastNotification"])

        def wrapper(res):
            return Notification(self.client, res["id"])

        return wrapWith(res, wrapper)

    @property
    def variables(self):
        from .variable import ThingVariablesList
        return ThingVariablesList(self.client, self.id)

    async def keep_online(self):
        async for _ in self.client.subscription_root.keepOnline(self._id):
            pass


class EnvironmentThingList:
    def __init__(self, client, environmentId):
        self.client = client
        self.environmentId = environmentId
        self.current = 0
        self._filter = "{}"

    def filter(self, _filter):
        self._filter = get_representation(_filter)
        return self

    def __len__(self):
        res = self.client.query(
            '{environment(id:"%s"){thingCount(filter:%s)}}' % (self.environmentId, self._filter))
        return res["environment"]["thingCount"]

    def __getitem__(self, i):
        if isinstance(i, int):
            res = self.client.query(
                '{environment(id:"%s"){things(limit:1, offset:%d, filter:%s){id}}}' % (self.environmentId, i, self._filter))
            if len(res["environment"]["things"]) != 1:
                raise IndexError()
            return Thing(self.client, res["environment"]["things"][0]["id"])
        elif isinstance(i, slice):
            start, end, _ = i.indices(len(self))
            res = self.client.query(
                '{environment(id:"%s"){things(offset:%d, limit:%d, filter:%s){id}}}' % (self.environmentId, start, end-start, self._filter))
            return [Thing(self.client, thing["id"]) for thing in res["environment"]["things"]]
        else:
            raise TypeError("Unexpected type {} passed as index".format(i))

    def __iter__(self):
        return self

    def __next__(self):
        res = self.client.query(
            '{environment(id:"%s"){things(limit:1, offset:%d, filter:%s){id}}}' % (self.environmentId, self.current, self._filter))

        if len(res["environment", "things"]) != 1:
            raise StopIteration

        self.current += 1
        return Thing(self.client, res["environment"]["things"][0]["id"])

    def next(self):
        return self.__next__()


class DeveloperThingList:
    def __init__(self, client, userId):
        self.client = client
        self.current = 0
        self._filter = "{}"
        self.userId = userId

    def filter(self, _filter):
        self._filter = get_representation(_filter)
        return self

    def __len__(self):
        res = self.client.query(
            '{user(id:%s){developerThingCount(filter:%s)}}' % (self.userId, self._filter))
        return res["user"]["developerThingCount"]

    def __getitem__(self, i):
        if isinstance(i, int):
            res = self.client.query(
                '{user(id:%s){developerThings(limit:1, offset:%d, filter:%s){id}}}' % (self.userId, i, self._filter))
            if len(res["user"]["developerThings"]) != 1:
                raise IndexError()
            return Thing(self.client, res["user"]["developerThings"][0]["id"])
        elif isinstance(i, slice):
            start, end, _ = i.indices(len(self))
            res = self.client.query(
                '{user(id:%s){developerThings(offset:%d, limit:%d, filter:%s){id}}}' % (self.userId, start, end-start, self._filter))
            return [Thing(self.client, thing["id"]) for thing in res["user"]["developerThings"]]
        else:
            raise TypeError("Unexpected type {} passed as index".format(i))

    def __iter__(self):
        return self

    def __next__(self):
        res = self.client.query(
            '{user(id:%s){developerThings(limit:1, offset:%d, filter:%s){id}}}' % (self.userId, self.current, self._filter))

        if len(res["user", "developerThings"]) != 1:
            raise StopIteration

        self.current += 1
        return Thing(self.client, res["user"]["developerThings"][0]["id"])

    def next(self):
        return self.__next__()




from aiodataloader import DataLoader
from igloo.models.utils import wrapWith


class FloatSeriesVariableLoader(DataLoader):
    def __init__(self, client, id):
        super().__init__()
        self.client = client
        self._id = id

    async def batch_load_fn(self, keys):
        fields = " ".join(set(keys))
        res = await self.client.query('{floatSeriesVariable(id:"%s"){%s}}' % (self._id, fields), keys=["floatSeriesVariable"])

        # if fetching object the key will be the first part of the field
        # e.g. when fetching thing{id} the result is in the thing key
        resolvedValues = [res[key.split("{")[0]] for key in keys]

        return resolvedValues


class FloatSeriesVariable:
    def __init__(self, client, id):
        self.client = client
        self._id = id
        self.loader = FloatSeriesVariableLoader(client, id)

    @property
    def id(self):
        return self._id

    @property
    def last_node(self):
        if self.client.asyncio:
            res = self.loader.load("lastNode{id}")
        else:
            res = self.client.query('{floatSeriesVariable(id:"%s"){lastNode{id}}}' % self._id, keys=[
                "floatSeriesVariable", "lastNode"])

        def wrapper(res):
            from .float_series_node import FloatSeriesNode
            return FloatSeriesNode(self.client, res["id"])

        return wrapWith(res, wrapper)

    @property
    def nodes(self):
        from .float_series_node import FloatSeriesNodeList
        return FloatSeriesNodeList(self.client, self._id)

    @property
    def name(self):
        if self.client.asyncio:
            return self.loader.load("name")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){name}}' % self._id, keys=[
                "floatSeriesVariable", "name"])

    @name.setter
    def name(self, newName):
        self.client.mutation(
            'mutation{floatSeriesVariable(id:"%s", name:"%s"){id}}' % (self._id, newName), asyncio=False)

    @property
    def developer_only(self):
        if self.client.asyncio:
            return self.loader.load("developerOnly")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){developerOnly}}' % self._id, keys=[
                "floatSeriesVariable", "developerOnly"])

    @developer_only.setter
    def developer_only(self, newValue):
        self.client.mutation(
            'mutation{floatSeriesVariable(id:"%s", developerOnly:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def hidden(self):
        if self.client.asyncio:
            return self.loader.load("hidden")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){hidden}}' % self._id, keys=[
                "floatSeriesVariable", "hidden"])

    @hidden.setter
    def hidden(self, newValue):
        self.client.mutation(
            'mutation{floatSeriesVariable(id:"%s", hidden:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def index(self):
        if self.client.asyncio:
            return self.loader.load("index")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){index}}' % self._id, keys=[
                "floatSeriesVariable", "index"])

    @index.setter
    def index(self, newValue):
        self.client.mutation(
            'mutation{floatSeriesVariable(id:"%s", index:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def shown_nodes(self):
        if self.client.asyncio:
            return self.loader.load("index")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){shownNodes}}' % self._id, keys=[
                "floatSeriesVariable", "shownNodes"])

    @shown_nodes.setter
    def shown_nodes(self, newValue):
        self.client.mutation(
            'mutation{floatSeriesVariable(id:"%s", shownNodes:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def stored_nodes(self):
        if self.client.asyncio:
            return self.loader.load("index")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){storedNodes}}' % self._id, keys=[
                "floatSeriesVariable", "storedNodes"])

    @stored_nodes.setter
    def stored_nodes(self, newValue):
        self.client.mutation(
            'mutation{floatSeriesVariable(id:"%s", storedNodes:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def my_role(self):
        if self.client.asyncio:
            return self.loader.load("myRole")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){myRole}}' % self._id, keys=[
                "floatSeriesVariable", "myRole"])

    @property
    def createdAt(self):
        if self.client.asyncio:
            return self.loader.load("createdAt")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){createdAt}}' % self._id, keys=[
                "floatSeriesVariable", "createdAt"])

    @property
    def updatedAt(self):
        if self.client.asyncio:
            return self.loader.load("updatedAt")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){updatedAt}}' % self._id, keys=[
                "floatSeriesVariable", "updatedAt"])

    async def _async_load_thing(self):
        id = await self.loader.load("thing{id}")["id"]
        from .thing import Thing
        return Thing(self.client, id)

    @property
    def thing(self):
        if self.client.asyncio:
            return self._async_load_thing()
        else:
            id = self.client.query('{floatSeriesVariable(id:"%s"){thing{id}}}' % self._id, keys=[
                "floatSeriesVariable", "thing", "id"])

            from .thing import Thing
            return Thing(self.client, id)

    @property
    def unit_of_measurement(self):
        if self.client.asyncio:
            return self.loader.load("unitOfMeasurement")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){unitOfMeasurement}}' % self._id, keys=[
                "floatSeriesVariable", "unitOfMeasurement"])

    @unit_of_measurement.setter
    def unit_of_measurement(self, newValue):
        self.client.mutation(
            'mutation{floatSeriesVariable(id:"%s", unitOfMeasurement:"%s"){id}}' % (self._id, newValue), asyncio=False)

    @property
    def precision(self):
        if self.client.asyncio:
            return self.loader.load("precision")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){precision}}' % self._id, keys=[
                "floatSeriesVariable", "precision"])

    @precision.setter
    def precision(self, newValue):
        self.client.mutation(
            'mutation{floatSeriesVariable(id:"%s", precision:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def min(self):
        if self.client.asyncio:
            return self.loader.load("min")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){min}}' % self._id, keys=[
                "floatSeriesVariable", "min"])

    @min.setter
    def min(self, newValue):
        self.client.mutation(
            'mutation{floatSeriesVariable(id:"%s", min:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def max(self):
        if self.client.asyncio:
            return self.loader.load("max")
        else:
            return self.client.query('{floatSeriesVariable(id:"%s"){max}}' % self._id, keys=[
                "floatSeriesVariable", "max"])

    @max.setter
    def max(self, newValue):
        self.client.mutation(
            'mutation{floatSeriesVariable(id:"%s", max:%s){id}}' % (self._id, newValue), asyncio=False)




from aiodataloader import DataLoader
from igloo.models.utils import wrapWith


class CategorySeriesVariableLoader(DataLoader):
    def __init__(self, client, id):
        super().__init__()
        self.client = client
        self._id = id

    async def batch_load_fn(self, keys):
        fields = " ".join(set(keys))
        res = await self.client.query('{categorySeriesVariable(id:"%s"){%s}}' % (self._id, fields), keys=["categorySeriesVariable"])

        # if fetching object the key will be the first part of the field
        # e.g. when fetching thing{id} the result is in the thing key
        resolvedValues = [res[key.split("{")[0]] for key in keys]

        return resolvedValues


class CategorySeriesVariable:
    def __init__(self, client, id):
        self.client = client
        self._id = id
        self.loader = CategorySeriesVariableLoader(client, id)

    @property
    def id(self):
        return self._id

    @property
    def last_node(self):
        if self.client.asyncio:
            res = self.loader.load("lastNode{id}")
        else:
            res = self.client.query('{categorySeriesVariable(id:"%s"){lastNode{id}}}' % self._id, keys=[
                "categorySeriesVariable", "lastNode"])

        def wrapper(res):
            from .category_series_node import CategorySeriesNode
            return CategorySeriesNode(self.client, res["id"])

        return wrapWith(res, wrapper)

    @property
    def nodes(self):
        from .category_series_node import CategorySeriesNodeList
        return CategorySeriesNodeList(self.client, self.id)

    @property
    def name(self):
        if self.client.asyncio:
            return self.loader.load("name")
        else:
            return self.client.query('{categorySeriesVariable(id:"%s"){name}}' % self._id, keys=[
                "categorySeriesVariable", "name"])

    @name.setter
    def name(self, newName):
        self.client.mutation(
            'mutation{categorySeriesVariable(id:"%s", name:"%s"){id}}' % (self._id, newName), asyncio=False)

    @property
    def developer_only(self):
        if self.client.asyncio:
            return self.loader.load("developerOnly")
        else:
            return self.client.query('{categorySeriesVariable(id:"%s"){developerOnly}}' % self._id, keys=[
                "categorySeriesVariable", "developerOnly"])

    @developer_only.setter
    def developer_only(self, newValue):
        self.client.mutation(
            'mutation{categorySeriesVariable(id:"%s", developerOnly:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def hidden(self):
        if self.client.asyncio:
            return self.loader.load("hidden")
        else:
            return self.client.query('{categorySeriesVariable(id:"%s"){hidden}}' % self._id, keys=[
                "categorySeriesVariable", "hidden"])

    @hidden.setter
    def hidden(self, newValue):
        self.client.mutation(
            'mutation{categorySeriesVariable(id:"%s", hidden:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def shown_nodes(self):
        if self.client.asyncio:
            return self.loader.load("shownNodes")
        else:
            return self.client.query('{categorySeriesVariable(id:"%s"){shownNodes}}' % self._id, keys=[
                "categorySeriesVariable", "shownNodes"])

    @shown_nodes.setter
    def shown_nodes(self, newValue):
        self.client.mutation(
            'mutation{categorySeriesVariable(id:"%s", shownNodes:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def stored_nodes(self):
        if self.client.asyncio:
            return self.loader.load("storedNodes")
        else:
            return self.client.query('{categorySeriesVariable(id:"%s"){storedNodes}}' % self._id, keys=[
                "categorySeriesVariable", "storedNodes"])

    @stored_nodes.setter
    def stored_nodes(self, newValue):
        self.client.mutation(
            'mutation{categorySeriesVariable(id:"%s", storedNodes:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def index(self):
        if self.client.asyncio:
            return self.loader.load("index")
        else:
            return self.client.query('{categorySeriesVariable(id:"%s"){index}}' % self._id, keys=[
                "categorySeriesVariable", "index"])

    @index.setter
    def index(self, newValue):
        self.client.mutation(
            'mutation{categorySeriesVariable(id:"%s", index:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def my_role(self):
        if self.client.asyncio:
            return self.loader.load("myRole")
        else:
            return self.client.query('{categorySeriesVariable(id:"%s"){myRole}}' % self._id, keys=[
                "categorySeriesVariable", "myRole"])

    @property
    def createdAt(self):
        if self.client.asyncio:
            return self.loader.load("createdAt")
        else:
            return self.client.query('{categorySeriesVariable(id:"%s"){createdAt}}' % self._id, keys=[
                "categorySeriesVariable", "createdAt"])

    @property
    def updatedAt(self):
        if self.client.asyncio:
            return self.loader.load("updatedAt")
        else:
            return self.client.query('{categorySeriesVariable(id:"%s"){updatedAt}}' % self._id, keys=[
                "categorySeriesVariable", "updatedAt"])

    async def _async_load_thing(self):
        id = await self.loader.load("thing{id}")["id"]
        from .thing import Thing
        return Thing(self.client, id)

    @property
    def thing(self):
        if self.client.asyncio:
            return self._async_load_thing()
        else:
            id = self.client.query('{categorySeriesVariable(id:"%s"){thing{id}}}' % self._id, keys=[
                "categorySeriesVariable", "thing", "id"])

            from .thing import Thing
            return Thing(self.client, id)

    @property
    def allowed_values(self):
        if self.client.asyncio:
            return self.loader.load("allowedValues")
        else:
            return self.client.query('{categorySeriesVariable(id:"%s"){allowedValues}}' % self._id, keys=[
                "categorySeriesVariable", "allowedValues"])

    @allowed_values.setter
    def allowed_values(self, newValue):
        self.client.mutation(
            'mutation{categorySeriesVariable(id:"%s", allowedValues:%s){id}}' % (self._id, str(newValue)), asyncio=False)




from aiodataloader import DataLoader


class FloatVariableLoader(DataLoader):
    def __init__(self, client, id):
        super().__init__()
        self.client = client
        self._id = id

    async def batch_load_fn(self, keys):
        fields = " ".join(set(keys))
        res = await self.client.query('{floatVariable(id:"%s"){%s}}' % (self._id, fields), keys=["floatVariable"])

        # if fetching object the key will be the first part of the field
        # e.g. when fetching thing{id} the result is in the thing key
        resolvedValues = [res[key.split("{")[0]] for key in keys]

        return resolvedValues


class FloatVariable:
    def __init__(self, client, id):
        self.client = client
        self._id = id
        self.loader = FloatVariableLoader(client, id)

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        if self.client.asyncio:
            return self.loader.load("name")
        else:
            return self.client.query('{floatVariable(id:"%s"){name}}' % self._id, keys=[
                "floatVariable", "name"])

    @name.setter
    def name(self, newName):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", name:"%s"){id}}' % (self._id, newName), asyncio=False)

    @property
    def private(self):
        if self.client.asyncio:
            return self.loader.load("private")
        else:
            return self.client.query('{floatVariable(id:"%s"){private}}' % self._id, keys=[
                "floatVariable", "private"])

    @private.setter
    def private(self, newValue):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", private:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def hidden(self):
        if self.client.asyncio:
            return self.loader.load("hidden")
        else:
            return self.client.query('{floatVariable(id:"%s"){hidden}}' % self._id, keys=[
                "floatVariable", "hidden"])

    @hidden.setter
    def hidden(self, newValue):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", hidden:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def index(self):
        if self.client.asyncio:
            return self.loader.load("index")
        else:
            return self.client.query('{floatVariable(id:"%s"){index}}' % self._id, keys=[
                "floatVariable", "index"])

    @index.setter
    def index(self, newValue):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", index:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def my_role(self):
        if self.client.asyncio:
            return self.loader.load("myRole")
        else:
            return self.client.query('{floatVariable(id:"%s"){myRole}}' % self._id, keys=[
                "floatVariable", "myRole"])

    @property
    def createdAt(self):
        if self.client.asyncio:
            return self.loader.load("createdAt")
        else:
            return self.client.query('{floatVariable(id:"%s"){createdAt}}' % self._id, keys=[
                "floatVariable", "createdAt"])

    @property
    def updatedAt(self):
        if self.client.asyncio:
            return self.loader.load("updatedAt")
        else:
            return self.client.query('{floatVariable(id:"%s"){updatedAt}}' % self._id, keys=[
                "floatVariable", "updatedAt"])

    async def _async_load_thing(self):
        id = await self.loader.load("thing{id}")["id"]

        from .thing import Thing
        return Thing(self.client, id)

    @property
    def thing(self):
        if self.client.asyncio:
            return self._async_load_thing()
        else:
            id = self.client.query('{floatVariable(id:"%s"){thing{id}}}' % self._id, keys=[
                "floatVariable", "thing", "id"])

            from .thing import Thing
            return Thing(self.client, id)

    @property
    def permission(self):
        if self.client.asyncio:
            return self.loader.load("permission")
        else:
            return self.client.query('{floatVariable(id:"%s"){permission}}' % self._id, keys=[
                "floatVariable", "permission"])

    @permission.setter
    def permission(self, newValue):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", permission:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def developer_only(self):
        if self.client.asyncio:
            return self.loader.load("developerOnly")
        else:
            return self.client.query('{floatVariable(id:"%s"){developerOnly}}' % self._id, keys=[
                "floatVariable", "developerOnly"])

    @developer_only.setter
    def developer_only(self, newValue):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", developerOnly:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def value(self):
        if self.client.asyncio:
            return self.loader.load("value")
        else:
            return self.client.query('{floatVariable(id:"%s"){value}}' % self._id, keys=[
                "floatVariable", "value"])

    @value.setter
    def value(self, newValue):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", value:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def precision(self):
        if self.client.asyncio:
            return self.loader.load("precision")
        else:
            return self.client.query('{floatVariable(id:"%s"){precision}}' % self._id, keys=[
                "floatVariable", "precision"])

    @precision.setter
    def precision(self, newValue):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", precision:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def min(self):
        if self.client.asyncio:
            return self.loader.load("min")
        else:
            return self.client.query('{floatVariable(id:"%s"){min}}' % self._id, keys=[
                "floatVariable", "min"])

    @min.setter
    def min(self, newValue):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", min:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def max(self):
        if self.client.asyncio:
            return self.loader.load("max")
        else:
            return self.client.query('{floatVariable(id:"%s"){max}}' % self._id, keys=[
                "floatVariable", "max"])

    @max.setter
    def max(self, newValue):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", max:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def allowed_values(self):
        if self.client.asyncio:
            return self.loader.load("allowedValues")
        else:
            return self.client.query('{floatVariable(id:"%s"){allowedValues}}' % self._id, keys=[
                "floatVariable", "allowedValues"])

    @allowed_values.setter
    def allowed_values(self, newValue):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", allowedValues:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def unit_of_measurement(self):
        if self.client.asyncio:
            return self.loader.load("unitOfMeasurement")
        else:
            return self.client.query('{floatVariable(id:"%s"){unitOfMeasurement}}' % self._id, keys=[
                "floatVariable", "unitOfMeasurement"])

    @unit_of_measurement.setter
    def unit_of_measurement(self, newValue):
        self.client.mutation(
            'mutation{floatVariable(id:"%s", unitOfMeasurement:"%s"){id}}' % (self._id, newValue), asyncio=False)



# proigloogrammatically generated file
from igloo.models.user import User
from igloo.models.access_token import AccessToken
from igloo.models.pending_share import PendingShare
from igloo.models.pending_transfer import PendingTransfer
from igloo.models.environment import Environment
from igloo.models.thing import Thing
from igloo.models.variable import Variable
from igloo.models.float_variable import FloatVariable
from igloo.models.notification import Notification
from igloo.models.boolean_variable import BooleanVariable
from igloo.models.string_variable import StringVariable
from igloo.models.float_series_variable import FloatSeriesVariable
from igloo.models.category_series_variable import CategorySeriesVariable
from igloo.models.category_series_node import CategorySeriesNode
from igloo.models.file_variable import FileVariable
from igloo.models.float_series_node import FloatSeriesNode
from igloo.utils import undefined, parse_arg


class SubscriptionRoot:
    def __init__(self, client):
        self.client = client

    async def thing_created(self, environment_id=undefined):
        environmentId_arg = parse_arg("environmentId", environment_id)

        async for data in self.client.subscribe(('subscription{thingCreated(%s){id}}' % (environmentId_arg)).replace('()', '')):
            yield Thing(self.client, data["thingCreated"]["id"])

    async def thing_claimed(self, environment_id=undefined, id=undefined):
        environmentId_arg = parse_arg(
            "environmentId", environment_id)
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{thingClaimed(%s%s){id}}' % (environmentId_arg, id_arg)).replace('()', '')):
            yield Thing(self.client, data["thingClaimed"]["id"])

    async def environment_created(self):
        async for data in self.client.subscribe(('subscription{environmentCreated(){id}}' % ()).replace('()', '')):
            yield Environment(self.client, data["environmentCreated"]["id"])

    async def variable_created(self, thing_id=undefined, hidden=undefined):
        thingId_arg = parse_arg("thingId", thing_id)
        hidden_arg = parse_arg("hidden", hidden)

        async for data in self.client.subscribe(('subscription{variableCreated(%s%s){id __typename}}' % (thingId_arg, hidden_arg)).replace('()', '')):
            yield Variable(self.client, data["variableCreated"]["id"], data["variableCreated"]["__typename"])

    async def float_series_node_created(self, series_id=undefined):
        seriesId_arg = parse_arg("seriesId", series_id)

        async for data in self.client.subscribe(('subscription{floatSeriesNodeCreated(%s){id}}' % (seriesId_arg)).replace('()', '')):
            yield FloatSeriesNode(self.client, data["floatSeriesNodeCreated"]["id"])

    async def category_series_node_created(self, series_id=undefined):
        seriesId_arg = parse_arg("seriesId", series_id)

        async for data in self.client.subscribe(('subscription{categorySeriesNodeCreated(%s){id}}' % (seriesId_arg)).replace('()', '')):
            yield CategorySeriesNode(self.client, data["categorySeriesNodeCreated"]["id"])

    async def access_token_created(self):
        async for data in self.client.subscribe(('subscription{accessTokenCreated(){id}}' % ()).replace('()', '')):
            yield AccessToken(self.client, data["accessTokenCreated"]["id"])

    async def notification_created(self):
        async for data in self.client.subscribe(('subscription{notificationCreated(){id}}' % ()).replace('()', '')):
            yield Notification(self.client, data["notificationCreated"]["id"])

    async def thing_moved(self, environment_id=undefined, id=undefined):
        environmentId_arg = parse_arg("environmentId", environment_id)
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{thingMoved(%s%s){id}}' % (environmentId_arg, id_arg)).replace('()', '')):
            yield Thing(self.client, data["thingMoved"]["id"])

    async def pending_share_created(self, ):
        async for data in self.client.subscribe(('subscription{pendingShareCreated(){id}}' % ()).replace('()', '')):
            yield PendingShare(self.client, data["pendingShareCreated"]["id"])

    async def pending_share_updated(self, ):
        async for data in self.client.subscribe(('subscription{pendingShareUpdated(){id}}' % ()).replace('()', '')):
            yield PendingShare(self.client, data["pendingShareUpdated"]["id"])

    async def pending_share_accepted(self, ):
        async for data in self.client.subscribe(('subscription{pendingShareAccepted(){id sender receiver role environment}}' % ()).replace('()', '')):
            yield data["pendingShareAccepted"]

    async def pending_share_declined(self, ):
        async for data in self.client.subscribe(('subscription{pendingShareDeclined()}' % ()).replace('()', '')):
            yield data["pendingShareDeclined"]

    async def pending_share_revoked(self, ):
        async for data in self.client.subscribe(('subscription{pendingShareRevoked()}' % ()).replace('()', '')):
            yield data["pendingShareRevoked"]

    async def user_left_environment(self, environment_id=undefined, user_id=undefined):
        environmentId_arg = parse_arg("environmentId", environment_id)
        userId_arg = parse_arg("userId", user_id)
        async for data in self.client.subscribe(('subscription{userLeftEnvironment(%s%s){environment{id} user{id}}}' % (environmentId_arg, userId_arg)).replace('()', '')):
            res = data["userLeftEnvironment"]
            res["environment"] = Environment(
                self.client, res["environment"]["id"])
            res["user"] = User(
                self.client, res["user"]["id"])

            yield res

    async def pending_transfer_created(self, ):
        async for data in self.client.subscribe(('subscription{pendingTransferCreated(){id}}' % ()).replace('()', '')):
            yield PendingTransfer(self.client, data["pendingTransferCreated"]["id"])

    async def pending_transfer_accepted(self):
        async for data in self.client.subscribe(('subscription{pendingTransferAccepted(){sender{id} receiver{id} environment{id} id}}').replace('()', '')):
            res = data["pendingTransferAccepted"]
            res["sender"] = User(
                self.client, res["sender"]["id"])
            res["receiver"] = User(
                self.client, res["receiver"]["id"])
            res["environment"] = Environment(
                self.client, res["environment"]["id"])

            yield res

    async def pending_transfer_declined(self, ):
        async for data in self.client.subscribe(('subscription{pendingTransferDeclined()}' % ()).replace('()', '')):
            yield data["pendingTransferDeclined"]

    async def pending_transfer_revoked(self, ):
        async for data in self.client.subscribe(('subscription{pendingTransferRevoked()}' % ()).replace('()', '')):
            yield data["pendingTransferRevoked"]

    async def user_updated(self, id=undefined, email=undefined):
        id_arg = parse_arg("id", id)
        email_arg = parse_arg("email", email)

        async for data in self.client.subscribe(('subscription{userUpdated(%s%s){id}}' % (id_arg, email_arg)).replace('()', '')):
            yield User(self.client, data["userUpdated"]["id"])

    async def thing_updated(self, environment_id=undefined, id=undefined):
        environmentId_arg = parse_arg("environmentId", environment_id)
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{thingUpdated(%s%s){id}}' % (environmentId_arg, id_arg)).replace('()', '')):
            yield Thing(self.client, data["thingUpdated"]["id"])

    async def environment_updated(self, id=undefined):
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{environmentUpdated(%s){id}}' % (id_arg)).replace('()', '')):
            yield Environment(self.client, data["environmentUpdated"]["id"])

    async def variable_updated(self, thing_id=undefined, id=undefined, hidden=undefined):
        thingId_arg = parse_arg("thingId", thing_id)
        id_arg = parse_arg("id", id)
        hidden_arg = parse_arg("hidden", hidden)

        async for data in self.client.subscribe(('subscription{variableUpdated(%s%s%s){id __typename}}' % (thingId_arg, id_arg, hidden_arg)).replace('()', '')):
            yield Variable(self.client, data["variableUpdated"]["id"], data["variableUpdated"]["__typename"])

    async def float_series_node_updated(self, series_id=undefined, id=undefined):
        seriesId_arg = parse_arg("seriesId", series_id)
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{floatSeriesNodeUpdated(%s%s){id}}' % (seriesId_arg, id_arg)).replace('()', '')):
            yield FloatSeriesNode(self.client, data["floatSeriesNodeUpdated"]["id"])

    async def category_series_node_updated(self, series_id=undefined, id=undefined):
        seriesId_arg = parse_arg("seriesId", series_id)
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{categorySeriesNodeUpdated(%s%s){id}}' % (seriesId_arg, id_arg)).replace('()', '')):
            yield CategorySeriesNode(self.client, data["categorySeriesNodeUpdated"]["id"])

    async def notification_updated(self, thing_id=undefined, id=undefined):
        thingId_arg = parse_arg("thingId", thing_id)
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{notificationUpdated(%s%s){id}}' % (thingId_arg, id_arg)).replace('()', '')):
            yield Notification(self.client, data["notificationUpdated"]["id"])

    async def variable_deleted(self, thing_id=undefined, id=undefined, hidden=undefined):
        thingId_arg = parse_arg("thingId", thing_id)
        id_arg = parse_arg("id", id)
        hidden_arg = parse_arg("hidden", hidden)

        async for data in self.client.subscribe(('subscription{variableDeleted(%s%s%s)}' % (thingId_arg, id_arg, hidden_arg)).replace('()', '')):
            yield data["variableDeleted"]

    async def float_series_node_deleted(self, series_id=undefined, id=undefined):
        seriesId_arg = parse_arg("seriesId", series_id)
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{floatSeriesNodeDeleted(%s%s)}' % (seriesId_arg, id_arg)).replace('()', '')):
            yield data["floatSeriesNodeDeleted"]

    async def category_series_node_deleted(self, series_id=undefined, id=undefined):
        seriesId_arg = parse_arg("seriesId", series_id)
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{categorySeriesNodeDeleted(%s%s)}' % (seriesId_arg, id_arg)).replace('()', '')):
            yield data["categorySeriesNodeDeleted"]

    async def thing_deleted(self, environment_id=undefined, id=undefined):
        environmentId_arg = parse_arg("environmentId", environment_id)
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{thingDeleted(%s%s)}' % (environmentId_arg, id_arg)).replace('()', '')):
            yield data["thingDeleted"]

    async def thing_unclaimed(self, environment_id=undefined, id=undefined):
        environmentId_arg = parse_arg("environmentId", environment_id)
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{thingUnclaimed(%s%s)}' % (environmentId_arg, id_arg)).replace('()', '')):
            yield data["thingUnclaimed"]

    async def environment_deleted(self, id=undefined):
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{environmentDeleted(%s)}' % (id_arg)).replace('()', '')):
            yield data["environmentDeleted"]

    async def user_deleted(self, id=undefined, email=undefined):
        id_arg = parse_arg("id", id)
        email_arg = parse_arg("id", email)

        async for data in self.client.subscribe(('subscription{userDeleted(%s%s)}' % (id_arg, email_arg)).replace('()', '')):
            yield data["userDeleted"]

    async def access_token_deleted(self, ):
        async for data in self.client.subscribe(('subscription{accessTokenDeleted()}' % ()).replace('()', '')):
            yield data["accessTokenDeleted"]

    async def notification_deleted(self, thing_id=undefined, id=undefined):
        thingId_arg = parse_arg("thingId", thing_id)
        id_arg = parse_arg("id", id)

        async for data in self.client.subscribe(('subscription{notificationDeleted(%s%s)}' % (thingId_arg, id_arg)).replace('()', '')):
            yield data["notificationDeleted"]

    async def keep_online(self, thing_id):
        thingId_arg = parse_arg("thingId", thing_id)

        async for data in self.client.subscribe(('subscription{keepOnline(%s)}' % (thingId_arg)).replace('()', '')):
            yield data["keepOnline"]



from igloo.models.user import User
from igloo.models.access_token import AccessToken
from igloo.models.pending_share import PendingShare
from igloo.models.environment import Environment
from igloo.models.thing import Thing
from igloo.models.float_variable import FloatVariable
from igloo.models.variable import Variable
from igloo.models.pending_transfer import PendingTransfer
from igloo.models.notification import Notification
from igloo.models.boolean_variable import BooleanVariable
from igloo.models.string_variable import StringVariable
from igloo.models.float_series_variable import FloatSeriesVariable
from igloo.models.category_series_variable import CategorySeriesVariable
from igloo.models.category_series_node import CategorySeriesNode
from igloo.models.file_variable import FileVariable
from igloo.models.float_series_node import FloatSeriesNode
from igloo.utils import undefined, parse_arg


class QueryRoot:
    def __init__(self, client):
        self.client = client

    @property
    def user(self, id=None, email=None):
        return User(self.client, id=id, email=email)

    def environment(self, id):
        return Environment(self.client, id)

    def thing(self, id):
        return Thing(self.client, id)

    def variable(self, id):
        return Variable(self.client, id)

    def float_variable(self, id):
        return FloatVariable(self.client, id)

    def string_variable(self, id):
        return StringVariable(self.client, id)

    def boolean_variable(self, id):
        return BooleanVariable(self.client, id)

    def file_variable(self, id):
        return FileVariable(self.client, id)

    def float_series_variable(self, id):
        return FloatSeriesVariable(self.client, id)

    def category_series_variable(self, id):
        return CategorySeriesVariable(self.client, id)

    def pending_share(self, id):
        return PendingShare(self.client, id)

    def pending_transfer(self, id):
        return PendingTransfer(self.client, id)

    def access_token(self, id):
        return AccessToken(self.client, id)

    def notification(self, id):
        return Notification(self.client, id)

    def float_series_node(self, id):
        return FloatSeriesNode(self.client, id)

    def category_series_node(self, id):
        return CategorySeriesNode(self.client, id)

    def get_new_totp_secret(self):
        return self.client.query("{getNewTotpSecret{secret,qrCode}}", keys=["getNewTotpSecret"])

    def verify_password(self, password, email=undefined):
        email_arg = parse_arg("email", email)
        return self.client.query('{verifyPassword(password:"%s" %s)}' % (password, email_arg), keys=["verifyPassword"])

    def verify_totp(self, code, email=undefined):
        email_arg = parse_arg("email", email)
        return self.client.query('{verifyTotp(code:"%s" %s)}' % (code, email_arg), keys=["verifyTotp"])

    def metadata(self):
        return self.client.query(
            """
            {
                metadata{
                    STANDARD_PLAN_MONTHLY_PRICE
                    STANDARD_PLAN_YEARLY_PRICE
                    EXTRA_STORAGE_MONTHLY_PRICE
                    EXTRA_THROUGHPUT_MONTHLY_PRICE
                    EXTRA_STORAGE_YEARLY_PRICE
                    EXTRA_THROUGHPUT_YEARLY_PRICE
                    CUSTOM_APPS_MONTHLY_PRICE
                    CUSTOM_APPS_YEARLY_PRICE
                }
            }
            """, keys=["metadata"])




from aiodataloader import DataLoader
from igloo.models.utils import wrapWith


class PendingTransferLoader(DataLoader):
    def __init__(self, client, id):
        super().__init__()
        self.client = client
        self._id = id

    async def batch_load_fn(self, keys):
        fields = " ".join(set(keys))
        res = await self.client.query('{pendingTransfer(id:"%s"){%s}}' % (self._id, fields), keys=["pendingTransfer"])

        # if fetching object the key will be the first part of the field
        # e.g. when fetching thing{id} the result is in the thing key
        resolvedValues = [res[key.split("{")[0]] for key in keys]

        return resolvedValues


class PendingTransfer:
    def __init__(self, client, id):
        self.client = client
        self._id = id
        self.loader = PendingTransferLoader(client, id)

    @property
    def id(self):
        return self._id

    @property
    def sender(self):
        if self.client.asyncio:
            res = self.loader.load("sender{id}")
        else:
            res = self.client.query('{pendingTransfer(id:"%s"){sender{id}}}' % self._id, keys=[
                "pendingTransfer", "sender"])

        def wrapper(res):
            from .user import User
            return User(self.client, res["id"])

        return wrapWith(res, wrapper)

    @property
    def receiver(self):
        if self.client.asyncio:
            res = self.loader.load("receiver{id}")
        else:
            res = self.client.query('{pendingTransfer(id:"%s"){receiver{id}}}' % self._id, keys=[
                "pendingTransfer", "receiver"])

        def wrapper(res):
            from .user import User
            return User(self.client, res["id"])

        return wrapWith(res, wrapper)

    @property
    def environment(self):
        if self.client.asyncio:
            res = self.loader.load("environment{id}")
        else:
            res = self.client.query('{pendingTransfer(id:"%s"){environment{id}}}' % self._id, keys=[
                "pendingTransfer", "environment"])

        def wrapper(res):
            from .environment import Environment
            return Environment(self.client, res["id"])

        return wrapWith(res, wrapper)


class UserPendingTransferList:
    def __init__(self, client, userId):
        self.client = client
        self.current = 0
        self.userId = userId

    def __len__(self):
        res = self.client.query('{user(id:%s){pendingTransferCount}}' % (self.userId), keys=[
                                "user", "pendingTransferCount"])
        return res

    def __getitem__(self, i):
        if isinstance(i, int):
            res = self.client.query(
                '{user(id:%s){pendingTransfers(limit:1, offset:%d){id}}}' % (self.userId, i))
            if len(res["user"]["pendingTransfers"]) != 1:
                raise IndexError()
            return PendingTransfer(self.client, res["user"]["pendingTransfers"][0]["id"])
        elif isinstance(i, slice):
            start, end, _ = i.indices(len(self))
            res = self.client.query(
                '{user(id:%s){pendingTransfers(offset:%d, limit:%d){id}}}' % (self.userId, start, end-start))
            return [PendingTransfer(self.client, ownerChange["id"]) for ownerChange in res["user"]["pendingTransfers"]]
        else:
            raise TypeError("Unexpected type {} passed as index".format(i))

    def __iter__(self):
        return self

    def __next__(self):
        res = self.client.query(
            '{user(id:%s){pendingTransfers(limit:1, offset:%d){id}}}' % (self.userId, self.current))

        if len(res["user"]["pendingTransfers"]) != 1:
            raise StopIteration

        self.current += 1
        return PendingTransfer(self.client, res["user"]["pendingTransfers"][0]["id"])

    def next(self):
        return self.__next__()




from aiodataloader import DataLoader
from igloo.models.utils import wrapWith


class AccessTokenLoader(DataLoader):
    def __init__(self, client, id):
        super().__init__()
        self.client = client
        self._id = id

    async def batch_load_fn(self, keys):
        fields = " ".join(set(keys))
        res = await self.client.query('{accessToken(id:"%s"){%s}}' % (self._id, fields), keys=["accessToken"])

        # if fetching object the key will be the first part of the field
        # e.g. when fetching thing{id} the result is in the thing key
        resolvedValues = [res[key.split("{")[0]] for key in keys]

        return resolvedValues


class AccessToken:
    def __init__(self, client, id):
        self.client = client
        self._id = id
        self.loader = AccessTokenLoader(client, id)

    @property
    def id(self):
        return self._id

    @property
    def user(self):
        if self.client.asyncio:
            res = self.loader.load("user{id}")
        else:
            res = self.client.query('{accessToken(id:"%s"){user{id}}}' % self._id, keys=[
                "accessToken", "user"])

        from .user import User

        def wrapper(res):
            return User(self.client, res["id"])

        return wrapWith(res, wrapper)

    @property
    def name(self):
        if self.client.asyncio:
            return self.loader.load("name")
        else:
            return self.client.query('{accessToken(id:"%s"){name}}' % self._id, keys=[
                "accessToken", "name"])

    @property
    def createdAt(self):
        if self.client.asyncio:
            return self.loader.load("createdAt")
        else:
            return self.client.query('{accessToken(id:"%s"){createdAt}}' % self._id, keys=[
                "accessToken", "createdAt"])

    @property
    def updatedAt(self):
        if self.client.asyncio:
            return self.loader.load("updatedAt")
        else:
            return self.client.query('{accessToken(id:"%s"){updatedAt}}' % self._id, keys=[
                "accessToken", "updatedAt"])

    @property
    def lastUsed(self):
        if self.client.asyncio:
            return self.loader.load("lastUsed")
        else:
            return self.client.query('{accessToken(id:"%s"){lastUsed}}' % self._id, keys=[
                "accessToken", "lastUsed"])


class AccessTokenList:
    def __init__(self, client, userId):
        self.client = client
        self.current = 0
        self.userId = userId

    def __len__(self):
        res = self.client.query(
            '{user(id:%s){accessTokenCount}}' % self.userId)
        return res["user"]["accessTokenCount"]

    def __getitem__(self, i):
        if isinstance(i, int):
            res = self.client.query(
                '{user(id:%s){accessTokens(limit:1, offset:%d){id}}}' % (self.userId, i))
            if len(res["user"]["accessTokens"]) != 1:
                raise IndexError()
            return AccessToken(self.client, res["user"]["accessTokens"][0]["id"])
        elif isinstance(i, slice):
            start, end, _ = i.indices(len(self))
            res = self.client.query(
                '{user(id:%s){accessTokens(offset:%d, limit:%d){id}}}' % (self.userId, start, end-start))
            return [AccessToken(self.client, token["id"]) for token in res["user"]["accessTokens"]]
        else:
            raise TypeError("Unexpected type {} passed as index".format(i))

    def __iter__(self):
        return self

    def __next__(self):
        res = self.client.query(
            '{user(id:%s){accessTokens(limit:1, offset:%d){id}}}' % (self.userId, self.current))

        if len(res["user", "accessTokens"]) != 1:
            raise StopIteration

        self.current += 1
        return AccessToken(self.client, res["user"]["accessTokens"][0]["id"])

    def next(self):
        return self.__next__()



import setuptools

setuptools.setup(
    name='igloo-python',
    packages=['igloo', 'igloo.models'],
    version='1.0.2',
    license='MIT',
    description='Python client for igloo',
    author='Igloo Team',
    author_email='hello@igloo.ooo',
    url='https://github.com/IglooCloud/igloo_python',
    keywords=['iot', 'igloo'],
    install_requires=[
        'requests', 'asyncio', 'pathlib', 'websockets', 'aiodataloader', 'aiohttp'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development ',
        'License :: OSI Approved :: MIT License  ',
        'Programming Language :: Python :: 3',
    ],
)




from aiodataloader import DataLoader
from igloo.models.utils import wrapWith


class CategorySeriesNodeLoader(DataLoader):
    def __init__(self, client, id):
        super().__init__()
        self.client = client
        self._id = id

    async def batch_load_fn(self, keys):
        fields = " ".join(set(keys))
        res = await self.client.query('{categorySeriesNode(id:"%s"){%s}}' % (self._id, fields), keys=["categorySeriesNode"])

        # if fetching object the key will be the first part of the field
        # e.g. when fetching thing{id} the result is in the thing key
        resolvedValues = [res[key.split("{")[0]] for key in keys]

        return resolvedValues


class CategorySeriesNode:
    def __init__(self, client, id):
        self.client = client
        self._id = id
        self.loader = CategorySeriesNodeLoader(client, id)

    @property
    def id(self):
        return self._id

    @property
    def createdAt(self):
        if self.client.asyncio:
            return self.loader.load("createdAt")
        else:
            return self.client.query('{categorySeriesNode(id:"%s"){createdAt}}' % self._id, keys=[
                "categorySeriesNode", "createdAt"])

    @property
    def updatedAt(self):
        if self.client.asyncio:
            return self.loader.load("updatedAt")
        else:
            return self.client.query('{categorySeriesNode(id:"%s"){updatedAt}}' % self._id, keys=[
                "categorySeriesNode", "updatedAt"])

    @property
    def thing(self):
        if self.client.asyncio:
            res = self.loader.load("thing{id}")
        else:
            res = self.client.query('{categorySeriesNode(id:"%s"){thing{id}}}' % self._id, keys=[
                "categorySeriesNode", "thing"])

        def wrapper(res):
            from .thing import Thing
            return Thing(self.client, res["id"])

        return wrapWith(res, wrapper)

    @property
    def series(self):
        if self.client.asyncio:
            res = self.loader.load("series{id}")
        else:
            res = self.client.query('{categorySeriesNode(id:"%s"){series{id}}}' % self._id, keys=[
                "categorySeriesNode", "series"])

        def wrapper(res):
            from .category_series_variable import CategorySeriesVariable
            return CategorySeriesVariable(self.client, res["id"])

        return wrapWith(res, wrapper)

    @property
    def timestamp(self):
        if self.client.asyncio:
            return self.loader.load("timestamp")
        else:
            return self.client.query('{categorySeriesNode(id:"%s"){timestamp}}' % self._id, keys=[
                "categorySeriesNode", "timestamp"])

    @timestamp.setter
    def timestamp(self, newValue):
        self.client.mutation(
            'mutation{categorySeriesNode(id:"%s", timestamp:"%s"){id}}' % (self._id, newValue), asyncio=False)

    @property
    def value(self):
        if self.client.asyncio:
            return self.loader.load("value")
        else:
            return self.client.query('{categorySeriesNode(id:"%s"){value}}' % self._id, keys=[
                "categorySeriesNode", "value"])

    @value.setter
    def value(self, newValue):
        self.client.mutation(
            'mutation{categorySeriesNode(id:"%s", value:"%s"){id}}' % (self._id, newValue), asyncio=False)


class CategorySeriesNodeList:
    def __init__(self, client, seriesId):
        self.client = client
        self.seriesId = seriesId
        self.current = 0

    def __len__(self):
        res = self.client.query(
            '{categorySeriesVariable(id:"%s"){nodeCount}}' % self.seriesId)
        return res["categorySeriesVariable"]["nodeCount"]

    def __getitem__(self, i):
        if isinstance(i, int):
            res = self.client.query(
                '{categorySeriesVariable(id:"%s"){nodes(limit:1, offset:%d){id}}}' % (self.seriesId, i))
            if len(res["categorySeriesVariable"]["nodes"]) != 1:
                raise IndexError()
            return CategorySeriesNode(self.client, res["categorySeriesVariable"]["nodes"][0]["id"])
        elif isinstance(i, slice):
            start, end, _ = i.indices(len(self))
            res = self.client.query(
                '{categorySeriesVariable(id:"%s"){nodes(offset:%d, limit:%d){id}}}' % (self.seriesId, start, end-start))
            return [CategorySeriesNode(self.client, node["id"]) for node in res["categorySeriesVariable"]["nodes"]]
        else:
            raise TypeError("Unexpected type {} passed as index".format(i))

    def __iter__(self):
        return self

    def __next__(self):
        res = self.client.query(
            '{categorySeriesVariable(id:"%s"){nodes(limit:1, offset:%d){id}}}' % (self.seriesId, self.current))

        if len(res["categorySeriesVariable", "nodes"]) != 1:
            raise StopIteration

        self.current += 1
        return CategorySeriesNode(self.client, res["categorySeriesVariable"]["nodes"][0]["id"])

    def next(self):
        return self.__next__()



from igloo.models.user import User
from igloo.models.access_token import AccessToken
from igloo.models.pending_share import PendingShare
from igloo.models.environment import Environment
from igloo.models.thing import Thing
from igloo.models.variable import Variable
from igloo.models.float_variable import FloatVariable
from igloo.models.pending_transfer import PendingTransfer
from igloo.models.notification import Notification
from igloo.models.boolean_variable import BooleanVariable
from igloo.models.string_variable import StringVariable
from igloo.models.float_series_variable import FloatSeriesVariable
from igloo.models.category_series_variable import CategorySeriesVariable
from igloo.models.category_series_node import CategorySeriesNode
from igloo.models.file_variable import FileVariable
from igloo.models.float_series_node import FloatSeriesNode
from igloo.utils import parse_arg, undefined


async def _asyncWrapWith(res, wrapper_fn):
    result = await res
    return wrapper_fn(result["id"])


def wrapById(res, wrapper_fn):
    if isinstance(res, dict):
        return wrapper_fn(res["id"])
    else:
        return _asyncWrapWith(res, wrapper_fn)


def wrapWith(res, wrapper_fn):
    if isinstance(res, dict):
        return wrapper_fn(res)
    else:
        return _asyncWrapWith(res, wrapper_fn)


class MutationRoot:
    def __init__(self, client):
        self.client = client

    def sendConfirmationEmail(self, email, operation):
        email_arg = parse_arg("email", email)
        operation_arg = parse_arg("operation", operation, is_enum=True)

        return self.client.mutation('mutation{sendConfirmationEmail(%s%s)}' % (email_arg, operation_arg))["sendConfirmationEmail"]

    async def _wrapLogIn(self, res):
        resDict = await res
        resDict["user"] = User(self.client)
        return resDict

    def log_in(self, email, password, totp=undefined, private_cloud=undefined):
        email_arg = parse_arg("email", email)
        password_arg = parse_arg("password", password)
        totp_arg = parse_arg("totp", totp)
        privateCloud_arg = parse_arg("privateCloud", private_cloud)

        res = self.client.mutation('mutation{logIn(%s%s%s%s){user{id} token}}' % (email_arg,
                                                                                  password_arg,
                                                                                  totp_arg,
                                                                                  privateCloud_arg))["logIn"]

        if isinstance(res, dict):
            self.client.set_token(res["token"])
            res["user"] = User(self.client)
            return res
        else:
            return self._wrapLogIn(res)

    def create_access_token(self, name, password):
        name_arg = parse_arg("name", name)
        password_arg = parse_arg("password", password)
        return self.client.mutation('mutation{createAccessToken(%s%s)}' % (name_arg, password_arg))["createAccessToken"]

    def regenerate_access_token(self, id, password):
        id_arg = parse_arg("id", id)
        password_arg = parse_arg("password", password)

        return self.client.mutation('mutation{regenerateAccessToken(%s%s)}' % (id_arg, password_arg))["regenerateAccessToken"]

    def delete_access_token(self, id, password):
        id_arg = parse_arg("id", id)
        password_arg = parse_arg("password", password)

        return self.client.mutation('mutation{deleteAccessToken(%s%s)}' % (id_arg, password_arg))["deleteAccessToken"]

    async def _wrapSignUp(self, res):
        result = await res
        result["user"] = User(self.client)

        return result

    def sign_up(self, email, name, password, accept_privacy_policy_and_EULA, company_name=undefined, private_cloud=undefined):
        email_arg = parse_arg("email", email)
        name_arg = parse_arg("name", name)
        password_arg = parse_arg("password", password)
        companyName_arg = parse_arg("companyName", company_name)
        privateCloud_arg = parse_arg("privateCloud", private_cloud)
        acceptPrivacyPolicyAndEULA_arg = parse_arg(
            "acceptPrivacyPolicyAndEULA", accept_privacy_policy_and_EULA)

        res = self.client.mutation('mutation{signUp(%s%s%s%s%s%s){user{id} token}}' % (
            email_arg,
            name_arg,
            password_arg,
            companyName_arg,
            privateCloud_arg,
            acceptPrivacyPolicyAndEULA_arg))["signUp"]

        if isinstance(res, dict):
            res["user"] = User(self.client, res["user"]["id"])
            return res
        else:
            return self._wrapSignUp(res)

    def initiate_billing_setup(self):
        res = self.client.mutation("mutation{initiateBillingSetup }")[
            "initiateBillingSetup"]

        return res

    def update_billing_info(self, stripe_payment_method):
        stripePaymentMethod_arg = parse_arg(
            "stripePaymentMethod", stripe_payment_method)
        res = self.client.mutation("mutation{updateBillingInfo(%s) }" % stripePaymentMethod_arg)[
            "updateBillingInfo"]

        return res

    def confirm_payment_execution(self):
        res = self.client.mutation("mutation{confirmPaymentExecution }")[
            "confirmPaymentExecution"]

        return res

    def retry_payment(self):
        res = self.client.mutation("mutation{retryPayment }")[
            "retryPayment"]

        return res

    def change_billing_plan(self, billing_plan, billing_cycle, extra_storage, extra_throughput, custom_apps):
        billingPlan_arg = parse_arg("billingPlan", billing_plan, is_enum=True)
        billingCycle_arg = parse_arg(
            "billingCycle", billing_cycle, is_enum=True)
        extraStorage_arg = parse_arg("extraStorage", extra_storage)
        extraThroughput_arg = parse_arg("extraThroughput", extra_throughput)
        customApps_arg = parse_arg("customApps", custom_apps)

        res = self.client.mutation("mutation{changeBillingPlan(%s%s%s%s%s) }" % (
            billingPlan_arg,
            billingCycle_arg,
            extraStorage_arg,
            extraThroughput_arg,
            customApps_arg
        ))[
            "changeBillingPlan"]

        return res

    def change_password(self, new_password, old_password):
        new_password_arg = parse_arg("newPassword", new_password)
        old_password_arg = parse_arg("oldPassword", old_password)

        res = self.client.mutation(
            'mutation{changePassword(%s%s)}' % (new_password_arg, old_password_arg))["changePassword"]

        return res

    def set_totp(self, code, secret, password):
        code_arg = parse_arg("code", code)
        secret_arg = parse_arg("secret", secret)
        password_arg = parse_arg("password", password)
        return self.client.mutation('mutation{setTotp(%s%s%s)}' % (code_arg, secret_arg, password_arg))["setTotp"]

    def disable_totp(self, password):
        password_arg = parse_arg("password", password)
        return self.client.mutation('mutation{disableTotp(%s%s%s)}' % password_arg)["disableTotp"]

    def send_disable_totp_email(self, email, redirect_to):
        email_arg = parse_arg("email", email)
        redirect_to_arg = parse_arg("redirectTo", redirect_to, is_enum=True)
        return self.client.mutation('mutation{sendDisableTotpEmail(%s%s)}' % (email_arg, redirect_to_arg))["sendDisableTotpEmail"]

    def send_verification_email(self, redirect_to):
        redirect_to_arg = parse_arg("redirectTo", redirect_to, is_enum=True)

        return self.client.mutation('mutation{sendVerificationEmail(%s)}' % (redirect_to_arg))["sendVerificationEmail"]

    def send_password_recovery_email(self, email, redirect_to):
        email_arg = parse_arg("email", email)
        redirect_to_arg = parse_arg("redirectTo", redirect_to, is_enum=True)
        return self.client.mutation('mutation{sendPasswordRecoveryEmail(%s%s)}' % (email_arg, redirect_to_arg))["sendPasswordRecoveryEmail"]

    def reset_password(self, recovery_token, new_password):
        recovery_token_arg = parse_arg("recoveryToken", recovery_token)
        new_password_arg = parse_arg("newPassword", new_password)
        return self.client.mutation('mutation{resetPassword(%s%s)}' % (recovery_token_arg, new_password_arg))["resetPassword"]

    def share_environment(self, environment_id, role, email=undefined, user_id=undefined):
        environmentId_arg = parse_arg("environmentId", environment_id)
        role_arg = parse_arg("role", role, is_enum=True)
        email_arg = parse_arg("email", email)
        userId_arg = parse_arg("userId", user_id)
        res = self.client.mutation('mutation{shareEnvironment(%s%s%s%s){id}}' % (
            environmentId_arg, email_arg, userId_arg, role_arg))["shareEnvironment"]

        def wrapper(id):
            return PendingShare(self.client, id)

        return wrapById(res, wrapper)

    def pending_share(self, id, role):
        id_arg = parse_arg("id", id)
        role_arg = parse_arg("role", role, is_enum=True)

        res = self.client.mutation('mutation{pendingShare(%s%s){id}}' % (
            id_arg, role_arg))["pendingShare"]

        def wrapper(id):
            return PendingShare(self.client, id)

        return wrapById(res, wrapper)

    def revoke_pending_share(self, pending_share_id):
        pendingShareId_arg = parse_arg(
            "pendingShareId", pending_share_id)

        return self.client.mutation('mutation{revokePendingShare(%s)}' % (pendingShareId_arg))["revokePendingShare"]

    def accept_pending_share(self, pending_share_id):
        pendingShareId_arg = parse_arg(
            "pendingShareId", pending_share_id)

        res = self.client.mutation('mutation{acceptPendingShare(%s){sender{id} receiver{id} role environment{id}}}' % (
            pendingShareId_arg))["acceptPendingShare"]

        def wrapper(res):
            res["sender"] = User(self.client, res["sender"]["id"])
            res["receiver"] = User(self.client, res["receiver"]["id"])
            res["environment"] = Environment(
                self.client, res["environment"]["id"])

            return res

        return wrapWith(res, wrapper)

    def decline_pending_share(self, pending_share_id):
        pendingShareId_arg = parse_arg(
            "pendingShareId", pending_share_id)

        return self.client.mutation('mutation{declinePendingShare(%s)}' % (pendingShareId_arg))["declinePendingShare"]

    def stop_sharing_environment(self, environment_id, email=undefined, user_id=undefined):
        environmentId_arg = parse_arg("environmentId", environment_id)
        email_arg = parse_arg("email", email)
        userId_arg = parse_arg("userId", user_id)
        res = self.client.mutation('mutation{stopSharingEnvironment(%s%s%s){id}}' % (
            environmentId_arg, email_arg, userId_arg))["stopSharingEnvironment"]

        def wrapper(id):
            return Environment(self.client, id)

        return wrapById(res, wrapper)

    def leave_environment(self, environment_id):
        environmentId_arg = parse_arg("environmentId", environment_id)

        return self.client.mutation('mutation{leaveEnvironment(%s)}' % (environmentId_arg))["leaveEnvironment"]

    def transfer_environment(self, environment_id, email=undefined, user_id=undefined):
        environmentId_arg = parse_arg("environmentId", environment_id)
        email_arg = parse_arg("email", email)
        userId_arg = parse_arg("userId", user_id)
        res = self.client.mutation('mutation{transferEnvironment(%s%s%s){id}}' % (
            environmentId_arg, email_arg, userId_arg))["transferEnvironment"]

        def wrapper(id):
            return PendingTransfer(self.client, id)

        return wrapById(res, wrapper)

    def revoke_pending_transfer(self, pending_transfer_id):
        pending_transfer_id_arg = parse_arg(
            "pendingTransferId", pending_transfer_id)

        return self.client.mutation('mutation{revokePendingTransfer(%s)}' % (pending_transfer_id_arg))["revokePendingTransfer"]

    def accept_pending_transfer(self, pending_transfer_id):
        pending_transfer_id_arg = parse_arg(
            "pendingTransferId", pending_transfer_id)

        res = self.client.mutation('mutation{acceptPendingTransfer(%s){id sender{id} receiver{id} environment{id}}}' % (
            pending_transfer_id_arg))["acceptPendingTransfer"]

        def wrapper(res):
            res["sender"] = User(self.client, res["sender"]["id"])
            res["receiver"] = User(self.client, res["receiver"]["id"])
            res["environment"] = Environment(
                self.client, res["environment"]["id"])

            return res

        return wrapWith(res, wrapper)

    def decline_pending_transfer(self, pending_transfer_id):
        pending_transfer_id_arg = parse_arg(
            "pendingTransferId", pending_transfer_id)

        return self.client.mutation('mutation{declinePendingTransfer(%s)}' % (pending_transfer_id_arg))["declinePendingTransfer"]

    def change_role(self, environment_id, email, new_role):
        environmentId_arg = parse_arg("environmentId", environment_id)
        email_arg = parse_arg("email", email)
        newRole_arg = parse_arg("newRole", new_role)

        res = self.client.mutation('mutation{changeRole(%s%s%s){id}}' % (
            environmentId_arg, email_arg, newRole_arg))["changeRole"]

        def wrapper(id):
            return Environment(self.client, id)

        return wrapById(res, wrapper)

    def create_environment(self, name, picture=undefined, index=undefined, muted=undefined):
        name_arg = parse_arg("name", name)
        picture_arg = parse_arg("picture", picture, is_enum=True)
        index_arg = parse_arg("index", index)
        muted_arg = parse_arg("muted", muted)
        res = self.client.mutation('mutation{createEnvironment(%s%s%s%s){id}}' % (
            name_arg, picture_arg, index_arg, muted_arg))["createEnvironment"]

        def wrapper(id):
            return Environment(self.client, id)

        return wrapById(res, wrapper)

    def create_thing(self, type, firmware=undefined, battery_threshold=undefined, stored_notifications=undefined):
        type_arg = parse_arg("type", type)
        firmware_arg = parse_arg("firmware", firmware)
        battery_threshold_arg = parse_arg(
            "batteryThreshold", battery_threshold)
        stored_notifications_arg = parse_arg(
            "storedNotifications", stored_notifications)
        res = self.client.mutation('mutation{createThing(%s%s%s%s){id}}' % (
            type_arg, firmware_arg, battery_threshold_arg, stored_notifications_arg))["createThing"]

        # FIXME: if we choose to keep the createThingPayload implement it here
        def wrapper(id):
            return Thing(self.client, id)

        return wrapById(res, wrapper)

    def claim_thing(self, claim_code, name, environment_id, index=undefined, muted=undefined):
        claimCode_arg = parse_arg("claimCode", claim_code)
        name_arg = parse_arg("name", name)
        environmentId_arg = parse_arg("environmentId", environment_id)
        index_arg = parse_arg("index", index)
        muted_arg = parse_arg("muted", muted)
        res = self.client.mutation('mutation{claimThing(%s%s%s%s%s){id}}' % (
            claimCode_arg, name_arg, index_arg, environmentId_arg, muted_arg))["claimThing"]

        def wrapper(id):
            return Thing(self.client, id)

        return wrapById(res, wrapper)

    def create_notification(self, thing_id, content, date=undefined):
        thingId_arg = parse_arg("thingId", thing_id)
        content_arg = parse_arg("content", content)
        date_arg = parse_arg("date", date)
        res = self.client.mutation('mutation{createNotification(%s%s%s){id}}' % (
            thingId_arg, content_arg, date_arg))["createNotification"]

        def wrapper(id):
            return Notification(self.client, id)

        return wrapById(res, wrapper)

    def create_float_variable(self, permission, name, thing_id=undefined, developer_only=undefined, allowed_values=undefined, unit_of_measurement=undefined, value=undefined, precision=undefined, min=undefined, max=undefined, index=undefined):
        thingId_arg = parse_arg("thingId", thing_id)
        permission_arg = parse_arg("permission", permission, is_enum=True)
        name_arg = parse_arg("name", name)
        developer_only_arg = parse_arg("developerOnly", developer_only)
        allowed_values_arg = parse_arg("hidden", allowed_values)
        unitOfMeasurement_arg = parse_arg(
            "unitOfMeasurement", unit_of_measurement)
        value_arg = parse_arg("value", value)
        precision_arg = parse_arg("precision", precision)
        min_arg = parse_arg("min", min)
        max_arg = parse_arg("max", max)

        index_arg = parse_arg("index", index)
        res = self.client.mutation('mutation{createFloatVariable(%s%s%s%s%s%s%s%s%s%s%s){id}}' % (thingId_arg, permission_arg, allowed_values_arg, developer_only_arg,
                                                                                                  unitOfMeasurement_arg, value_arg, precision_arg, min_arg, max_arg, name_arg, index_arg))["createFloatVariable"]

        def wrapper(id):
            return FloatVariable(self.client, id)

        return wrapById(res, wrapper)

    def create_string_variable(self, permission, name, thing_id=undefined, developer_only=undefined, value=undefined, max_characters=undefined, allowed_values=undefined, index=undefined):
        thingId_arg = parse_arg("thingId", thing_id)
        permission_arg = parse_arg("permission", permission, is_enum=True)
        name_arg = parse_arg("name", name)
        developer_only_arg = parse_arg("developerOnly", developer_only)
        value_arg = parse_arg("value", value)
        maxChars_arg = parse_arg("maxChars", max_characters)

        allowedValues_arg = parse_arg("allowedValues", allowed_values)
        index_arg = parse_arg("index", index)
        res = self.client.mutation('mutation{createStringVariable(%s%s%s%s%s%s%s%s){id}}' % (
            thingId_arg, permission_arg, developer_only_arg, value_arg, maxChars_arg, name_arg, allowedValues_arg, index_arg))["createStringVariable"]

        def wrapper(id):
            return StringVariable(self.client, id)

        return wrapById(res, wrapper)

    def create_boolean_variable(self, permission, name, thingId=undefined, developer_only=undefined,  value=undefined, index=undefined):
        thingId_arg = parse_arg("thingId", thingId)
        permission_arg = parse_arg("permission", permission, is_enum=True)
        name_arg = parse_arg("name", name)
        developer_only_arg = parse_arg("developerOnly", developer_only)
        value_arg = parse_arg("value", value)

        index_arg = parse_arg("index", index)
        res = self.client.mutation('mutation{createBooleanVariable(%s%s%s%s%s%s){id}}' % (
            thingId_arg, permission_arg, developer_only_arg, value_arg, name_arg, index_arg))["createBooleanVariable"]

        def wrapper(id):
            return BooleanVariable(self.client, id)

        return wrapById(res, wrapper)

    def create_float_series_variable(self, name, shown_nodes, thing_id=undefined, developer_only=undefined, unit_of_measurement=undefined, precision=undefined, min=undefined, max=undefined, index=undefined, stored_nodes=undefined):
        thingId_arg = parse_arg("thingId", thing_id)
        name_arg = parse_arg("name", name)
        developer_only_arg = parse_arg("private", developer_only)
        unitOfMeasurement_arg = parse_arg(
            "unitOfMeasurement", unit_of_measurement)
        precision_arg = parse_arg("precision", precision)
        min_arg = parse_arg("min", min)
        max_arg = parse_arg("max", max)
        shown_nodes_arg = parse_arg("max", shown_nodes)
        stored_nodes_arg = parse_arg("max", stored_nodes)

        index_arg = parse_arg("index", index)
        res = self.client.mutation('mutation{createFloatSeriesVariable(%s%s%s%s%s%s%s%s%s%s){id}}' % (
            shown_nodes_arg, stored_nodes_arg, thingId_arg, developer_only_arg, unitOfMeasurement_arg, precision_arg, min_arg, max_arg, name_arg, index_arg))["createFloatSeriesVariable"]

        def wrapper(id):
            return FloatSeriesVariable(self.client, id)

        return wrapById(res, wrapper)

    def create_float_series_node(self, series_id, value=undefined, timestamp=undefined):
        seriesId_arg = parse_arg("seriesId", series_id)
        value_arg = parse_arg("value", value)
        timestamp_arg = parse_arg("timestamp", timestamp)
        res = self.client.mutation('mutation{createFloatSeriesNode(%s%s%s){id}}' % (
            seriesId_arg, timestamp_arg, value_arg))["createFloatSeriesNode"]

        def wrapper(id):
            return FloatSeriesNode(self.client, id)

        return wrapById(res, wrapper)

    def user(self,
             company_name=undefined,
             quiet_mode=undefined,
             name=undefined,
             language=undefined,
             vat_number=undefined,
             lenght_and_mass=undefined,
             temperature=undefined,
             date_format=undefined,
             time_format=undefined,
             password_change_email=undefined,
             shares_email=undefined,
             access_token_created_email=undefined,
             address_line1=undefined,
             address_line2=undefined,
             address_postal_code=undefined,
             address_city=undefined,
             address_state=undefined,
             address_country_or_territory=undefined
             ):

        company_name_arg = parse_arg("company_name", company_name)
        quiet_mode_arg = parse_arg("quiet_mode", quiet_mode)
        name_arg = parse_arg("name", name)
        language_arg = parse_arg("language", language)
        vat_number_arg = parse_arg("vat_number", vat_number)
        lenght_and_mass_arg = parse_arg(
            "lenght_and_mass", lenght_and_mass)
        temperature_arg = parse_arg("temperature", temperature)
        date_format_arg = parse_arg("date_format", date_format)
        time_format_arg = parse_arg("time_format", time_format)
        password_change_email_arg = parse_arg(
            "password_change_email", password_change_email)
        shares_email_arg = parse_arg("shares_email", shares_email)
        access_token_created_email_arg = parse_arg(
            "access_token_created_email", access_token_created_email)
        address_line1_arg = parse_arg("address_line1", address_line1)
        address_line2_arg = parse_arg("address_line2", address_line2)
        address_postal_code_arg = parse_arg(
            "address_postal_code", address_postal_code)
        address_city_arg = parse_arg("address_city", address_city)
        address_state_arg = parse_arg("address_state", address_state)
        address_country_or_territory_arg = parse_arg(
            "address_country_or_territory", address_country_or_territory)

        res = self.client.mutation('mutation{user(%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s){id}}' % (
            company_name_arg, quiet_mode_arg, name_arg, language_arg, vat_number_arg, lenght_and_mass_arg, temperature_arg, date_format_arg, time_format_arg, password_change_email_arg, shares_email_arg, access_token_created_email_arg, address_line1_arg, address_line2_arg, address_postal_code_arg, address_city_arg, address_state_arg, address_country_or_territory_arg
        ))["user"]

        def wrapper(id):
            return User(self.client)

        return wrapById(res, wrapper)

    def change_email(self, newEmail, password, redirect_to):
        newEmail_arg = parse_arg("newEmail", newEmail)
        password_arg = parse_arg("password", password)
        redirect_to_arg = parse_arg("redirectTo", redirect_to)

        return self.client.mutation('mutation{changeEmail(%s%s%s)}' % (newEmail_arg, password_arg, redirect_to_arg))["changeEmail"]

    def environment(self, id, name=undefined, picture=undefined, index=undefined, muted=undefined):
        id_arg = parse_arg("id", id)
        name_arg = parse_arg("name", name)
        picture_arg = parse_arg("picture", picture, is_enum=True)
        index_arg = parse_arg("index", index)
        muted_arg = parse_arg("muted", muted)
        res = self.client.mutation('mutation{environment(%s%s%s%s%s){id}}' % (
            id_arg, name_arg, picture_arg, index_arg, muted_arg))["environment"]

        def wrapper(id):
            return Environment(self.client, id)

        return wrapById(res, wrapper)

    def thing(self, id, type=undefined, name=undefined, index=undefined, signal_status=undefined, battery_status=undefined, battery_charging=undefined, battery_threshold=undefined, firmware=undefined, muted=undefined, starred=undefined, stored_notifications=undefined):
        id_arg = parse_arg("id", id)
        thingType_arg = parse_arg("type", type)
        name_arg = parse_arg("name", name)
        index_arg = parse_arg("index", index)
        signalStatus_arg = parse_arg("signalStatus", signal_status)
        batteryStatus_arg = parse_arg("batteryStatus", battery_status)
        batteryCharging_arg = parse_arg("batteryCharging", battery_charging)
        firmware_arg = parse_arg("firmware", firmware)
        muted_arg = parse_arg("muted", muted)
        starred_arg = parse_arg("starred", starred)
        battery_threshold_arg = parse_arg(
            "batteryThreshold", battery_threshold)
        stored_notifications_arg = parse_arg(
            "storedNotifications", stored_notifications)
        res = self.client.mutation('mutation{thing(%s%s%s%s%s%s%s%s%s%s%s%s){id}}' % (
            battery_threshold_arg, stored_notifications_arg, id_arg, thingType_arg, name_arg, index_arg, signalStatus_arg, batteryStatus_arg, batteryCharging_arg, firmware_arg, muted_arg, starred_arg))["thing"]

        def wrapper(id):
            return Thing(self.client, id)

        return wrapById(res, wrapper)

    def move_thing(self, thing_id, new_environment_id):
        thing_id_arg = parse_arg("thingId", thing_id)
        new_environment_id_arg = parse_arg(
            "newEnvironmentId", new_environment_id)

        res = self.client.mutation('mutation{moveThing(%s%s){id}}' % (
            thing_id_arg, new_environment_id_arg))["moveThing"]

        def wrapper(id):
            return Thing(self.client, id)

        return wrapById(res, wrapper)

    def value(self, id, developer_only=undefined, hidden=undefined, name=undefined, index=undefined):
        id_arg = parse_arg("id", id)
        developer_only_arg = parse_arg("developerOnly", developer_only)
        hidden_arg = parse_arg("hidden", hidden)

        name_arg = parse_arg("name", name)
        index_arg = parse_arg("index", index)
        res = self.client.mutation('mutation{value(%s%s%s%s%s){id __typename}}' % (
            id_arg, developer_only_arg, hidden_arg, name_arg, index_arg))["value"]

        def wrapper(res):
            return Value(self.client, res["id"], res["__typename"])

        return wrapWith(res, wrapper)

    def reset_online_state(self, thing_id):
        thingId_arg = parse_arg("thingId", thing_id)

        res = self.client.mutation('mutation{resetOnlineState(%s){id}}' % (thingId_arg))[
            "resetOnlineState"]

        def wrapper(id):
            return Thing(self.client, id)

        return wrapById(res, wrapper)

    def float_variable(self, id, permission=undefined, developer_only=undefined, hidden=undefined, unit_of_measurement=undefined, value=undefined, precision=undefined, min=undefined, max=undefined, name=undefined, index=undefined, allowed_values=undefined):
        id_arg = parse_arg("id", id)
        permission_arg = parse_arg("permission", permission, is_enum=True)
        developer_only_arg = parse_arg("developerOnly", developer_only)
        hidden_arg = parse_arg("hidden", hidden)
        unit_of_measurement_arg = parse_arg(
            "unitOfMeasurement", unit_of_measurement)
        value_arg = parse_arg("value", value)
        precision_arg = parse_arg("precision", precision)
        min_arg = parse_arg("min", min)
        max_arg = parse_arg("max", max)
        name_arg = parse_arg("name", name)
        allowed_values_arg = parse_arg("allowedValues", allowed_values)

        index_arg = parse_arg("index", index)
        res = self.client.mutation('mutation{floatVariable(%s%s%s%s%s%s%s%s%s%s%s%s){id}}' % (
            id_arg, permission_arg, developer_only_arg, allowed_values_arg, hidden_arg, unit_of_measurement_arg, value_arg, precision_arg, min_arg, max_arg, name_arg, index_arg))["floatVariable"]

        def wrapper(id):
            return FloatVariable(self.client, id)

        return wrapById(res, wrapper)

    def increment_float_variable(self, id, increment_by):
        id_arg = parse_arg("id", id)
        incrementBy_arg = parse_arg("incrementBy", increment_by)

        res = self.client.mutation('mutation{incrementFloatVariable(%s%s){id}}' % (
            id_arg, incrementBy_arg))["incrementFloatVariable"]

        def wrapper(id):
            return FloatVariable(self.client, id)

        return wrapById(res, wrapper)

    def string_variable(self, id, permission=undefined, developer_only=undefined, hidden=undefined, value=undefined, max_characters=undefined, name=undefined, allowed_values=undefined, index=undefined):
        id_arg = parse_arg("id", id)
        permission_arg = parse_arg("permission", permission, is_enum=True)
        private_arg = parse_arg("developerOnly", developer_only)
        hidden_arg = parse_arg("hidden", hidden)
        value_arg = parse_arg("value", value)
        maxChars_arg = parse_arg("maxCharacters", max_characters)
        name_arg = parse_arg("name", name)

        allowedValues_arg = parse_arg("allowedValues", allowed_values)
        index_arg = parse_arg("index", index)
        res = self.client.mutation('mutation{stringVariable(%s%s%s%s%s%s%s%s%s){id}}' % (
            id_arg, permission_arg, private_arg, hidden_arg, value_arg, maxChars_arg, name_arg, allowedValues_arg, index_arg))["stringVariable"]

        def wrapper(id):
            return StringVariable(self.client, id)

        return wrapById(res, wrapper)

    def boolean_variable(self, id, permission=undefined, developer_only=undefined, hidden=undefined, value=undefined, name=undefined, index=undefined):
        id_arg = parse_arg("id", id)
        permission_arg = parse_arg("permission", permission, is_enum=True)
        private_arg = parse_arg("developerOnly", developer_only)
        hidden_arg = parse_arg("hidden", hidden)
        value_arg = parse_arg("value", value)
        name_arg = parse_arg("name", name)

        index_arg = parse_arg("index", index)
        res = self.client.mutation('mutation{booleanVariable(%s%s%s%s%s%s%s){id}}' % (
            id_arg, permission_arg, private_arg, hidden_arg, value_arg, name_arg, index_arg))["booleanVariable"]

        def wrapper(id):
            return BooleanVariable(self.client, id)

        return wrapById(res, wrapper)

    def float_series_variable(self, id, developer_only=undefined, hidden=undefined, unit_of_measurement=undefined, precision=undefined, min=undefined, max=undefined, name=undefined, index=undefined, shown_nodes=undefined, stored_nodes=undefined):
        id_arg = parse_arg("id", id)
        private_arg = parse_arg("developerOnly", developer_only)
        hidden_arg = parse_arg("hidden", hidden)
        unitOfMeasurement_arg = parse_arg(
            "unitOfMeasurement", unit_of_measurement)
        precision_arg = parse_arg("precision", precision)
        min_arg = parse_arg("min", min)
        max_arg = parse_arg("max", max)
        name_arg = parse_arg("name", name)
        shown_nodes_arg = parse_arg("shownNodes", shown_nodes)
        stored_nodes_arg = parse_arg("storedNodes", stored_nodes)

        index_arg = parse_arg("index", index)
        res = self.client.mutation('mutation{floatSeriesVariable(%s%s%s%s%s%s%s%s%s%s%s){id}}' % (
            id_arg, private_arg, shown_nodes_arg, stored_nodes_arg, hidden_arg, unitOfMeasurement_arg, precision_arg, min_arg, max_arg, name_arg, index_arg))["floatSeriesVariable"]

        def wrapper(id):
            return FloatSeriesVariable(self.client, id)

        return wrapById(res, wrapper)

    def float_series_node(self, id, value=undefined, timestamp=undefined):
        id_arg = parse_arg("id", id)
        value_arg = parse_arg("value", value)
        timestamp_arg = parse_arg("timestamp", timestamp)
        res = self.client.mutation('mutation{floatSeriesNode(%s%s%s){id}}' % (
            id_arg, value_arg, timestamp_arg))["floatSeriesNode"]

        def wrapper(id):
            return FloatSeriesNode(self.client, id)

        return wrapById(res, wrapper)

    def category_series_variable(self, id, developer_only=undefined, hidden=undefined, name=undefined, allowed_values=undefined, index=undefined, shown_nodes=undefined, stored_nodes=undefined):
        id_arg = parse_arg("id", id)
        developer_only_arg = parse_arg("developerOnly", developer_only)
        hidden_arg = parse_arg("hidden", hidden)
        name_arg = parse_arg("name", name)
        shown_nodes_arg = parse_arg("shownNodes", shown_nodes)
        stored_nodes_arg = parse_arg("storedNodes", stored_nodes)

        allowedValues_arg = parse_arg("allowedValues", allowed_values)
        index_arg = parse_arg("index", index)
        res = self.client.mutation('mutation{categorySeriesVariable(%s%s%s%s%s%s%s%s){id}}' % (
            id_arg, developer_only_arg, shown_nodes_arg, stored_nodes_arg, hidden_arg, name_arg, allowedValues_arg, index_arg))["categorySeriesVariable"]

        def wrapper(id):
            return CategorySeriesVariable(self.client, id)

        return wrapById(res, wrapper)

    def category_series_node(self, id, value=undefined, timestamp=undefined):
        id_arg = parse_arg("id", id)
        value_arg = parse_arg("value", value)
        timestamp_arg = parse_arg("timestamp", timestamp)
        res = self.client.mutation('mutation{categorySeriesNode(%s%s%s){id}}' % (
            id_arg, value_arg, timestamp_arg))["categorySeriesNode"]

        def wrapper(id):
            return CategorySeriesNode(self.client, id)

        return wrapById(res, wrapper)

    def notification(self, id, content=undefined, read=undefined):
        id_arg = parse_arg("id", id)
        content_arg = parse_arg("content", content)
        read_arg = parse_arg("read", read)
        res = self.client.mutation('mutation{notification(%s%s%s){id}}' % (
            id_arg, content_arg, read_arg))["notification"]

        def wrapper(id):
            return Notification(self.client, id)

        return wrapById(res, wrapper)

    def delete_notification(self, id):
        id_arg = parse_arg("id", id)

        return self.client.mutation('mutation{deleteNotification(%s)}' % (id_arg))["deleteNotification"]

    def delete_variable(self, id):
        id_arg = parse_arg("id", id)

        return self.client.mutation('mutation{deleteVariable(%s)}' % (id_arg))["deleteVariable"]

    def delete_thing(self, id):
        id_arg = parse_arg("id", id)

        return self.client.mutation('mutation{deleteThing(%s)}' % (id_arg))["deleteThing"]

    def unclaimThing(self, id, reset):
        id_arg = parse_arg("id", id)
        reset_arg = parse_arg("reset", reset)

        return self.client.mutation('mutation{unclaimThing(%s%s){id}}' % (id_arg, reset_arg))["unclaimThing"]

    def delete_environment(self, id):
        id_arg = parse_arg("id", id)

        return self.client.mutation('mutation{deleteEnvironment(%s)}' % (id_arg))["deleteEnvironment"]

    def delete_user(self, ):

        return self.client.mutation('mutation{deleteUser()}' % ())["deleteUser"]

    def delete_float_series_node(self, id):
        id_arg = parse_arg("id", id)

        return self.client.mutation('mutation{deleteFloatSeriesNode(%s)}' % (id_arg))["deleteFloatSeriesNode"]

    def delete_category_series_node(self, id):
        id_arg = parse_arg("id", id)

        return self.client.mutation('mutation{deleteCategorySeriesNode(%s)}' % (id_arg))["deleteCategorySeriesNode"]



from datetime import datetime


class _Undefined:
    pass


UNDEFINED = undefined = _Undefined()


def get_from_dict(d, keys):
    if len(keys) == 0:
        return d
    else:
        return get_from_dict(d[keys[0]], keys[1:])


def parse_arg(arg_name, arg_value, is_enum=False):
    if isinstance(arg_value, _Undefined):
        return ""
    else:
        return "%s:%s," % (arg_name, get_representation(arg_value, is_enum=is_enum))


def get_representation(arg_value, is_enum=False):
    if isinstance(arg_value, _Undefined):
        return ""
    elif arg_value is None:
        return "null"
    elif isinstance(arg_value, dict):
        parsed_items = []
        for key, value in arg_value.items():
            parsed_items.append('"%s":%s' % (key, get_representation(value)))

        return "{%s}" % ",".join(parsed_items)
    elif isinstance(arg_value, list):
        parsed_items = []
        for value in arg_value:
            parsed_items.append(get_representation(value))

        return "[%s]" % ",".join(parsed_items)
    elif is_enum:
        return str(arg_value)
    elif isinstance(arg_value, str):
        return '"%s"' % arg_value
    elif isinstance(arg_value, bool):
        return "true" if arg_value else "false"
    elif isinstance(arg_value, datetime):
        return '"%s"' % arg_value.isoformat()
    else:
        return str(arg_value)



from igloo.models.utils import wrapWith
from igloo.utils import get_representation
from aiodataloader import DataLoader


class EnvironmentLoader(DataLoader):
    def __init__(self, client, id):
        super().__init__()
        self.client = client
        self._id = id

    async def batch_load_fn(self, keys):
        fields = " ".join(set(keys))
        res = await self.client.query('{environment(id:"%s"){%s}}' % (self._id, fields), keys=["environment"])

        resolvedValues = [res[key.split("{")[0]] for key in keys]

        return resolvedValues


class Environment:
    def __init__(self, client, id):
        self.client = client
        self._id = id
        self.loader = EnvironmentLoader(client, id)

    @property
    def id(self):
        return self._id

    @property
    def createdAt(self):
        if self.client.asyncio:
            return self.loader.load("createdAt")
        else:
            return self.client.query('{environment(id:"%s"){createdAt}}' % self._id, keys=[
                "environment", "createdAt"])

    @property
    def updatedAt(self):
        if self.client.asyncio:
            return self.loader.load("updatedAt")
        else:
            return self.client.query('{environment(id:"%s"){updatedAt}}' % self._id, keys=[
                "environment", "updatedAt"])

    @property
    def name(self):
        if self.client.asyncio:
            return self.loader.load("name")
        else:
            return self.client.query('{environment(id:"%s"){name}}' % self._id, keys=[
                "environment", "name"])

    @name.setter
    def name(self, newName):
        self.client.mutation(
            'mutation{environment(id:"%s", name:"%s"){id}}' % (self._id, newName), asyncio=False)

    @property
    def owner(self):
        if self.client.asyncio:
            res = self.loader.load("owner{id}")
        else:
            res = self.client.query('{environment(id:"%s"){owner{id}}}' % self._id, keys=[
                "environment", "owner"])

        def wrapper(res):
            from .user import User
            res = User(self.client, res["id"])

            return res

        return wrapWith(res, wrapper)

    @property
    def my_role(self):
        if self.client.asyncio:
            return self.loader.load("myRole")
        else:
            return self.client.query('{environment(id:"%s"){myRole}}' % self._id, keys=[
                "environment", "myRole"])

    @property
    def picture(self):
        if self.client.asyncio:
            return self.loader.load("picture")
        else:
            return self.client.query('{environment(id:"%s"){picture}}' % self._id, keys=[
                "environment", "picture"])

    @picture.setter
    def picture(self, newPicture):
        self.client.mutation(
            'mutation{environment(id:"%s", picture:"%s"){id}}' % (self._id, newPicture), asyncio=False)

    @property
    def unique_firmwares(self):
        if self.client.asyncio:
            return self.loader.load("uniqueFirmwares")
        else:
            return self.client.query('{environment(id:"%s"){uniqueFirmwares}}' % self._id, keys=[
                "environment", "uniqueFirmwares"])

    @property
    def index(self):
        if self.client.asyncio:
            return self.loader.load("index")
        else:
            return self.client.query('{environment(id:"%s"){index}}' % self._id, keys=[
                "environment", "index"])

    @index.setter
    def index(self, newIndex):
        self.client.mutation(
            'mutation{environment(id:"%s", index:%s){id}}' % (self._id, newIndex), asyncio=False)

    @property
    def muted(self):
        if self.client.asyncio:
            return self.loader.load("muted")
        else:
            return self.client.query('{environment(id:"%s"){muted}}' % self._id, keys=[
                "environment", "muted"])

    @muted.setter
    def muted(self, newMuted):
        self.client.mutation(
            'mutation{environment(id:"%s", muted:%s){id}}' % (self._id, "true" if newMuted else "false"), asyncio=False)

    @property
    def things(self):
        from .thing import EnvironmentThingList
        return EnvironmentThingList(self.client, self._id)

    @property
    def editors(self):
        from .user import EnvironmentEditorList
        return EnvironmentEditorList(self.client, self._id)

    @property
    def spectators(self):
        from .user import EnvironmentSpectatorList
        return EnvironmentSpectatorList(self.client, self._id)

    @property
    def pending_shares(self):
        from .pending_share import EnvironmentPendingShareList
        return EnvironmentPendingShareList(self.client, self._id)

    @property
    def pending_transfer(self):
        if self.client.asyncio:
            res = self.loader.load("pendingTransfer{id}")
        else:
            res = self.client.query('{environment(id:"%s"){pendingTransfer{id}}}' % self._id, keys=[
                "environment", "pendingTransfer"])

        def wrapper(res):
            from .pending_transfer import PendingTransfer
            res = PendingTransfer(self.client, res["id"])

            return res

        return wrapWith(res, wrapper)


class EnvironmentList:
    def __init__(self, client, userId):
        self.client = client
        self.current = 0
        self._filter = "{}"
        self.userId = userId

    def filter(self, _filter):
        self._filter = get_representation(_filter)
        return self

    def __len__(self):
        res = self.client.query('{user(id:%s){environmentCount(filter:%s)}}' % (self.userId, self._filter), keys=[
                                "user", "environmentCount"])
        return res

    def __getitem__(self, i):
        if isinstance(i, int):
            res = self.client.query(
                '{user(id:%s){environments(limit:1, offset:%d, filter:%s){id}}}' % (self.userId, i, self._filter))
            if len(res["user"]["environments"]) != 1:
                raise IndexError()
            return Environment(self.client, res["user"]["environments"][0]["id"])
        elif isinstance(i, slice):
            start, end, _ = i.indices(len(self))
            res = self.client.query(
                '{user(id:%s){environments(offset:%d, limit:%d, filter:%s){id}}}' % (self.userId, start, end-start, self._filter))
            return [Environment(self.client, environment["id"]) for environment in res["user"]["environments"]]
        else:
            raise TypeError("Unexpected type {} passed as index".format(i))

    def __iter__(self):
        return self

    def __next__(self):
        res = self.client.query(
            '{user(id:%s){environments(limit:1, offset:%d, filter:%s){id}}}' % (self.userId, self.current, self._filter))

        if len(res["user"]["environments"]) != 1:
            raise StopIteration

        self.current += 1
        return Environment(self.client, res["user"]["environments"][0]["id"])

    def next(self):
        return self.__next__()



from aiodataloader import DataLoader


class UserLoader(DataLoader):
    def __init__(self, client, id):
        super().__init__()
        self.client = client
        self._id = id

    async def batch_load_fn(self, keys):
        fields = " ".join(set(keys))
        res = await self.client.query('{user(id:"%s"){%s}}' % (self._id, fields), keys=["user"])

        resolvedValues = [res[key] for key in keys]

        return resolvedValues


class User:
    def __init__(self, client, id=None, email=None):
        self.client = client

        if id is None and email is None:
            self._id = self.client.query(
                '{user{id}}', keys=["user", "id"], asyncio=False)
        elif id is None:
            self._id = self.client.query(
                '{user(email:"%s"){id}}' % email, keys=["user", "id"], asyncio=False)
        else:
            self._id = id

        self.loader = UserLoader(client, self._id)

    @property
    def id(self):
        return self._id

    @property
    def email(self):
        if self.client.asyncio:
            return self.loader.load("email")
        else:
            return self.client.query('{user(id:"%s"){email}}' % self._id, keys=["user", "email"])

    @property
    def name(self):
        if self.client.asyncio:
            return self.loader.load("name")
        else:
            return self.client.query('{user(id:"%s"){name}}' % self._id, keys=["user", "name"])

    @name.setter
    def name(self, newName):
        self.client.mutation(
            'mutation{user(id:"%s", name:"%s"){id}}' % (self._id, newName), asyncio=False)

    @property
    def company_name(self):
        if self.client.asyncio:
            return self.loader.load("companyName")
        else:
            return self.client.query('{user(id:"%s"){companyName}}' % self._id, keys=["user", "companyName"])

    @company_name.setter
    def company_name(self, newName):
        self.client.mutation(
            'mutation{user(id:"%s", companyName:"%s"){id}}' % (self._id, newName), asyncio=False)

    @property
    def profile_icon_color(self):
        if self.client.asyncio:
            return self.loader.load("profileIconColor")
        else:
            return self.client.query('{user(id:"%s"){profileIconColor}}' % self._id,
                                     keys=["user", "profileIconColor"])

    @property
    def quiet_mode(self):
        if self.client.asyncio:
            return self.loader.load("quietMode")
        else:
            return self.client.query('{user(id:"%s"){quietMode}}' % self._id, keys=[
                "user", "quietMode"])

    @quiet_mode.setter
    def quiet_mode(self, newMode):
        self.client.mutation(
            'mutation{user(id:"%s",quietMode:%s){id}}' % (self._id, "true" if newMode else "false"), asyncio=False)

    @property
    def address_line1(self):
        if self.client.asyncio:
            return self.loader.load("addressLine1")
        else:
            return self.client.query('{user(id:"%s"){addressLine1}}' % self._id, keys=[
                "user", "addressLine1"])

    @address_line1.setter
    def address_line1(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", addressLine1:"%s"){id}}' % (self._id, newValue), asyncio=False)

    @property
    def address_line2(self):
        if self.client.asyncio:
            return self.loader.load("addressLine2")
        else:
            return self.client.query('{user(id:"%s"){addressLine2}}' % self._id, keys=[
                "user", "addressLine2"])

    @address_line2.setter
    def address_line2(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", addressLine2:"%s"){id}}' % (self._id, newValue), asyncio=False)

    @property
    def address_postal_code(self):
        if self.client.asyncio:
            return self.loader.load("addressPostalCode")
        else:
            return self.client.query('{user(id:"%s"){addressPostalCode}}' % self._id, keys=[
                "user", "addressPostalCode"])

    @address_postal_code.setter
    def address_postal_code(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", addressPostalCode:"%s"){id}}' % (self._id, newValue), asyncio=False)

    @property
    def address_city(self):
        if self.client.asyncio:
            return self.loader.load("addressCity")
        else:
            return self.client.query('{user(id:"%s"){addressCity}}' % self._id, keys=[
                "user", "addressCity"])

    @address_city.setter
    def address_city(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", addressCity:"%s"){id}}' % (self._id, newValue), asyncio=False)

    @property
    def address_state(self):
        if self.client.asyncio:
            return self.loader.load("addressState")
        else:
            return self.client.query('{user(id:"%s"){addressState}}' % self._id, keys=[
                "user", "addressState"])

    @address_state.setter
    def address_state(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", addressState:"%s"){id}}' % (self._id, newValue), asyncio=False)

    @property
    def address_country_or_territory(self):
        if self.client.asyncio:
            return self.loader.load("addressCountryOrTerritory")
        else:
            return self.client.query('{user(id:"%s"){addressCountryOrTerritory}}' % self._id, keys=[
                "user", "addressCountryOrTerritory"])

    @address_country_or_territory.setter
    def address_country_or_territory(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", addressCountryOrTerritory:"%s"){id}}' % (self._id, newValue), asyncio=False)

    @property
    def billing_plan(self):
        if self.client.asyncio:
            return self.loader.load("billingPlan")
        else:
            return self.client.query('{user(id:"%s"){billingPlan}}' % self._id, keys=[
                "user", "billingPlan"])

    @property
    def billing_cycle(self):
        if self.client.asyncio:
            return self.loader.load("billingCycle")
        else:
            return self.client.query('{user(id:"%s"){billingCycle}}' % self._id, keys=[
                "user", "billingCycle"])

    @property
    def billing_status(self):
        if self.client.asyncio:
            return self.loader.load("billingStatus")
        else:
            return self.client.query('{user(id:"%s"){billingStatus}}' % self._id, keys=[
                "user", "billingStatus"])

    @property
    def payment_intent_secret(self):
        if self.client.asyncio:
            return self.loader.load("paymentIntentSecret")
        else:
            return self.client.query('{user(id:"%s"){paymentIntentSecret}}' % self._id, keys=[
                "user", "paymentIntentSecret"])

    @property
    def card_last4_digits(self):
        if self.client.asyncio:
            return self.loader.load("cardLast4Digits")
        else:
            return self.client.query('{user(id:"%s"){cardLast4Digits}}' % self._id, keys=[
                "user", "cardLast4Digits"])

    @property
    def card_expiry_month(self):
        if self.client.asyncio:
            return self.loader.load("cardExpiryMonth")
        else:
            return self.client.query('{user(id:"%s"){cardExpiryMonth}}' % self._id, keys=[
                "user", "cardExpiryMonth"])

    @property
    def card_expiry_year(self):
        if self.client.asyncio:
            return self.loader.load("cardExpiryYear")
        else:
            return self.client.query('{user(id:"%s"){cardExpiryYear}}' % self._id, keys=[
                "user", "cardExpiryYear"])

    @property
    def vat_number(self):
        if self.client.asyncio:
            return self.loader.load("vatNumber")
        else:
            return self.client.query('{user(id:"%s"){vatNumber}}' % self._id, keys=[
                "user", "vatNumber"])

    @vat_number.setter
    def vat_number(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", vatNumber:"%s"){id}}' % (self._id, newValue), asyncio=False)

    @property
    def next_billing_date(self):
        if self.client.asyncio:
            return self.loader.load("nextBillingDate")
        else:
            return self.client.query('{user(id:"%s"){nextBillingDate}}' % self._id, keys=[
                "user", "nextBillingDate"])

    @property
    def billing_credit(self):
        if self.client.asyncio:
            return self.loader.load("billingCredit")
        else:
            return self.client.query('{user(id:"%s"){billingCredit}}' % self._id, keys=[
                "user", "billingCredit"])

    @property
    def extra_storage(self):
        if self.client.asyncio:
            return self.loader.load("extraStorage")
        else:
            return self.client.query('{user(id:"%s"){extraStorage}}' % self._id, keys=[
                "user", "extraStorage"])

    @property
    def extra_throughput(self):
        if self.client.asyncio:
            return self.loader.load("extraThroughput")
        else:
            return self.client.query('{user(id:"%s"){extraThroughput}}' % self._id, keys=[
                "user", "extraThroughput"])

    @property
    def max_storage(self):
        if self.client.asyncio:
            return self.loader.load("maxStorage")
        else:
            return self.client.query('{user(id:"%s"){maxStorage}}' % self._id, keys=[
                "user", "maxStorage"])

    @property
    def max_throughput(self):
        if self.client.asyncio:
            return self.loader.load("maxThroughput")
        else:
            return self.client.query('{user(id:"%s"){maxThroughput}}' % self._id, keys=[
                "user", "maxThroughput"])

    @property
    def used_storage(self):
        if self.client.asyncio:
            return self.loader.load("usedStorage")
        else:
            return self.client.query('{user(id:"%s"){usedStorage}}' % self._id, keys=[
                "user", "usedStorage"])

    @property
    def used_throughput(self):
        if self.client.asyncio:
            return self.loader.load("usedThroughput")
        else:
            return self.client.query('{user(id:"%s"){usedThroughput}}' % self._id, keys=[
                "user", "usedThroughput"])

    @property
    def custom_apps(self):
        if self.client.asyncio:
            return self.loader.load("customApps")
        else:
            return self.client.query('{user(id:"%s"){customApps}}' % self._id, keys=[
                "user", "customApps"])

    @property
    def email_is_verified(self):
        if self.client.asyncio:
            return self.loader.load("emailIsVerified")
        else:
            return self.client.query('{user(id:"%s"){emailIsVerified}}' % self._id, keys=[
                "user", "emailIsVerified"])

    @property
    def unique_developer_firmwares(self):
        if self.client.asyncio:
            return self.loader.load("uniqueDeveloperFirmwares")
        else:
            return self.client.query('{user(id:"%s"){uniqueDeveloperFirmwares}}' % self._id, keys=[
                "user", "uniqueDeveloperFirmwares"])

    @property
    def environments(self):
        from .environment import EnvironmentList
        return EnvironmentList(self.client, self.id)

    @property
    def pendingShares(self):
        from .pending_share import UserPendingShareList
        return UserPendingShareList(self.client, self.id)

    @property
    def pendingTransfers(self):
        from .pending_transfer import UserPendingTransferList
        return UserPendingTransferList(self.client, self.id)

    @property
    def developerThings(self):
        from .thing import DeveloperThingList
        return DeveloperThingList(self.client, self.id)

    @property
    def accessTokens(self):
        from .access_token import AccessTokenList
        return AccessTokenList(self.client, self.id)

    @property
    def totp_enabled(self):
        if self.client.asyncio:
            return self.loader.load("totpEnabled")
        else:
            return self.client.query('{user(id:"%s"){totpEnabled}}' % self._id, keys=[
                "user", "totpEnabled"])

    @property
    def length_and_mass(self):
        if self.client.asyncio:
            return self.loader.load("lengthAndMass")
        else:
            return self.client.query('{user(id:"%s"){lengthAndMass}}' % self._id, keys=[
                "user", "lengthAndMass"])

    @length_and_mass.setter
    def length_and_mass(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", lengthAndMass:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def temperature(self):
        if self.client.asyncio:
            return self.loader.load("temperature")
        else:
            return self.client.query('{user(id:"%s"){temperature}}' % self._id, keys=[
                "user", "temperature"])

    @temperature.setter
    def temperature(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", temperature:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def time_format(self):
        if self.client.asyncio:
            return self.loader.load("timeFormat")
        else:
            return self.client.query('{user(id:"%s"){timeFormat}}' % self._id, keys=[
                "user", "timeFormat"])

    @time_format.setter
    def time_format(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", timeFormat:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def date_format(self):
        if self.client.asyncio:
            return self.loader.load("dateFormat")
        else:
            return self.client.query('{user(id:"%s"){dateFormat}}' % self._id, keys=[
                "user", "dateFormat"])

    @date_format.setter
    def date_format(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", dateFormat:%s){id}}' % (self._id, newValue), asyncio=False)

    @property
    def language(self):
        if self.client.asyncio:
            return self.loader.load("language")
        else:
            return self.client.query('{user(id:"%s"){language}}' % self._id, keys=[
                "user", "language"])

    @language.setter
    def language(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", language:"%s"){id}}' % (self._id, newValue), asyncio=False)

    @property
    def password_change_email(self):
        if self.client.asyncio:
            return self.loader.load("passwordChangeEmail")
        else:
            return self.client.query('{user(id:"%s"){passwordChangeEmail}}' % self._id, keys=[
                "user", "passwordChangeEmail"])

    @password_change_email.setter
    def password_change_email(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", passwordChangeEmail:%s){id}}' % (self._id, "true" if newValue == True else "false"), asyncio=False)

    @property
    def shares_email(self):
        if self.client.asyncio:
            return self.loader.load("sharesEmail")
        else:
            return self.client.query('{user(id:"%s"){sharesEmail}}' % self._id, keys=[
                "user", "sharesEmail"])

    @shares_email.setter
    def shares_email(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", sharesEmail:%s){id}}' % (self._id, "true" if newValue == True else "false"), asyncio=False)

    @property
    def access_token_created_email(self):
        if self.client.asyncio:
            return self.loader.load("accessTokenCreatedEmail")
        else:
            return self.client.query('{user(id:"%s"){accessTokenCreatedEmail}}' % self._id, keys=[
                "user", "accessTokenCreatedEmail"])

    @access_token_created_email.setter
    def access_token_created_email(self, newValue):
        self.client.mutation(
            'mutation{user(id:"%s", accessTokenCreatedEmail:%s){id}}' % (self._id, "true" if newValue == True else "false"), asyncio=False)

    @property
    def business_pricing(self):
        return self.client.query('{user(id:"%s"){businessPricing{id maxStorage maxThroughput price}}}' % self._id, keys=[
            "user", "businessPricing"])


class EnvironmentEditorList:
    def __init__(self, client, environmentId):
        self.client = client
        self.environmentId = environmentId
        self.current = 0

    def __len__(self):
        res = self.client.query(
            '{environment(id:"%s"){editorCount}}' % self.environmentId)
        return res["environment"]["editorCount"]

    def __getitem__(self, i):
        if isinstance(i, int):
            res = self.client.query(
                '{environment(id:"%s"){editors(limit:1, offset:%d){id}}}' % (self.environmentId, i))
            if len(res["environment"]["editors"]) != 1:
                raise IndexError()
            return User(self.client, res["environment"]["editors"][0]["id"])
        elif isinstance(i, slice):
            start, end, _ = i.indices(len(self))
            res = self.client.query(
                '{environment(id:"%s"){editors(offset:%d, limit:%d){id}}}' % (self.environmentId, start, end-start))
            return [User(self.client, user["id"]) for user in res["environment"]["editors"]]
        else:
            raise TypeError("Unexpected type {} passed as index".format(i))

    def __iter__(self):
        return self

    def __next__(self):
        res = self.client.query(
            '{environment(id:"%s"){editors(limit:1, offset:%d){id}}}' % (self.environmentId, self.current))

        if len(res["environment", "editors"]) != 1:
            raise StopIteration

        self.current += 1
        return User(self.client, res["environment"]["editors"][0]["id"])

    def next(self):
        return self.__next__()


class EnvironmentSpectatorList:
    def __init__(self, client, environmentId):
        self.client = client
        self.environmentId = environmentId
        self.current = 0

    def __len__(self):
        res = self.client.query(
            '{environment(id:"%s"){spectatorCount}}' % self.environmentId)
        return res["environment"]["spectatorCount"]

    def __getitem__(self, i):
        if isinstance(i, int):
            res = self.client.query(
                '{environment(id:"%s"){spectators(limit:1, offset:%d){id}}}' % (self.environmentId, i))
            if len(res["environment"]["spectators"]) != 1:
                raise IndexError()
            return User(self.client, res["environment"]["spectators"][0]["id"])
        elif isinstance(i, slice):
            start, end, _ = i.indices(len(self))
            res = self.client.query(
                '{environment(id:"%s"){spectators(offset:%d, limit:%d){id}}}' % (self.environmentId, start, end-start))
            return [User(self.client, user["id"]) for user in res["environment"]["spectators"]]
        else:
            raise TypeError("Unexpected type {} passed as index".format(i))

    def __iter__(self):
        return self

    def __next__(self):
        res = self.client.query(
            '{environment(id:"%s"){spectators(limit:1, offset:%d){id}}}' % (self.environmentId, self.current))

        if len(res["environment", "spectators"]) != 1:
            raise StopIteration

        self.current += 1
        return User(self.client, res["environment"]["spectators"][0]["id"])

    def next(self):
        return self.__next__()



import requests
import asyncio
import pathlib
import websockets
import json
from igloo.mutations import MutationRoot
from igloo.subscriptions import SubscriptionRoot
from igloo.utils import get_from_dict, _Undefined, UNDEFINED, undefined
from aiohttp import ClientSession
from .query import QueryRoot
import asyncio

host = "api.igloo.ooo"
url = "https://{}/graphql".format(host)


class GraphQLException(Exception):
    pass


def exponential_backoff():
    yield 1
    yield 5

    while True:
        yield 20


class Client:
    def __init__(self, token, asyncio=False):
        self.token = token
        self.session = ClientSession()
        self.asyncio = asyncio

    def set_token(self, newToken):
        self.token = newToken

    def __del__(self):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.__close__())

    async def __close__(self):
        await self.session.close()

    @property
    def query_root(self):
        return QueryRoot(self)

    @property
    def mutation_root(self):
        return MutationRoot(self)

    @property
    def subscription_root(self):
        return SubscriptionRoot(self)

    async def __async_query(self, query, variables=None, keys=[]):
        payload = {"query": query}
        if variables != None:
            payload["variables"] = variables

        headers = {
            'content-type': "application/json",
            'authorization': "Bearer " + self.token
        }

        async with self.session.post(url, data=json.dumps(payload), headers=headers) as response:
            parsedRes = await response.json()

        if "errors" in parsedRes.keys():
            raise GraphQLException(parsedRes["errors"][0]["message"])

        return get_from_dict(parsedRes, ["data", *keys])

    def __sync_query(self, query, variables=None, keys=[]):
        payload = {"query": query}
        if variables != None:
            payload["variables"] = variables

        headers = {
            'content-type': "application/json",
            'authorization': "Bearer " + self.token
        }

        response = requests.request(
            "POST", url, data=json.dumps(payload), headers=headers)

        parsedRes = json.loads(response.text)
        if "errors" in parsedRes.keys():
            raise GraphQLException(parsedRes["errors"][0]["message"])

        return get_from_dict(parsedRes, ["data", *keys])

    def query(self, query, variables=None, keys=[], asyncio=None):
        if asyncio == False or (asyncio is None and not self.asyncio):
            return self.__sync_query(query, variables=variables, keys=keys)
        else:
            return self.__async_query(query, variables=variables, keys=keys)

    mutation = query

    async def _subscribe(self, query):
        async with websockets.connect(
                'wss://{}/subscriptions'.format(host), ssl=True, subprotocols=["graphql-ws"]) as websocket:
            await websocket.send('{"type":"connection_init","payload":{"Authorization":"Bearer %s"}}' % (self.token))

            res = await websocket.recv()
            if json.loads(res)["type"] != "connection_ack":
                raise Exception("failed to connect")

            # TODO: escape better the query
            listen_query_message = '{"id":"1","type":"start","payload":{"query":"%s","variables":null}}' % (
                query.replace('"', '\\"').replace('\n', '\\n')
            )
            await websocket.send(listen_query_message)
            while True:
                response = await websocket.recv()
                parsedResponse = json.loads(response)
                if parsedResponse["type"] == "data":
                    if "errors" in parsedResponse["payload"].keys():
                        raise GraphQLException(
                            parsedResponse["payload"]["errors"][0]["message"])
                    else:
                        yield parsedResponse["payload"]["data"]

    async def subscribe(self, query, autoreconnect=True):
        for backoff in exponential_backoff():
            try:
                async for res in self._subscribe(query):
                    yield res
            except GraphQLException:
                raise
            except Exception:
                if not autoreconnect:
                    raise
                await asyncio.sleep(backoff)
